# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""ACL Generator base class."""

import copy
import datetime
import hashlib
import re
import string
from typing import Optional, Union

from absl import logging

from aerleon.lib import policy


# generic error class
class Error(Exception):
    """Base error class."""


class NoPlatformPolicyError(Error):
    """Raised when a policy is received that doesn't support this platform."""


class UnknownIcmpTypeError(Error):
    """Raised when we see an unknown icmp-type."""


class MismatchIcmpInetError(Error):
    """Raised when mistmatch between icmp/icmpv6 and inet/inet6."""


class EstablishedError(Error):
    """Raised when a term has established option with inappropriate protocol."""


class UnsupportedAFError(Error):
    """Raised when provided an unsupported address family."""


class DuplicateTermError(Error):
    """Raised when duplication of term names are detected."""


class UnsupportedFilterError(Error):
    """Raised when we see an inappropriate filter."""


class UnsupportedTargetOptionError(Error):
    """Raised when a filter has an impermissible default action specified."""


class TermNameTooLongError(Error):
    """Raised when term named can not be abbreviated."""


class Term:
    """Generic framework for a generator Term."""

    ICMP_TYPE = policy.Term.ICMP_TYPE
    # http://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml
    PROTO_MAP = {
        'hopopt': 0,
        'icmp': 1,
        'igmp': 2,
        'ggp': 3,
        'ipip': 4,
        'tcp': 6,
        'egp': 8,
        'igp': 9,
        'udp': 17,
        'rdp': 27,
        'ipv6': 41,
        'ipv6-route': 43,
        'fragment': 44,
        'rsvp': 46,
        'gre': 47,
        'esp': 50,
        'ah': 51,
        'icmpv6': 58,
        'ipv6-nonxt': 59,
        'ipv6-opts': 60,
        'ospf': 89,
        'pim': 103,
        'vrrp': 112,
        'l2tp': 115,
        'sctp': 132,
        'udplite': 136,
        'all': -1,  # Used for GCE default deny, do not use in pol file.
    }
    AF_MAP = {
        'inet': 4,
        'inet6': 6,
        'bridge': 4,
        'ethernet-switching': 4,
    }
    # if this doesn't exist, output includes v4 & v6
    # These protos are always expressed as numbers instead of name
    #  due to inconsistencies on the end platform's name-to-number
    #  mapping.
    ALWAYS_PROTO_NUM = ['ipip']
    # provide flipped key/value dicts
    PROTO_MAP_BY_NUMBER = {v: k for (k, v) in PROTO_MAP.items()}
    AF_MAP_BY_NUMBER = {v: k for (k, v) in AF_MAP.items()}

    NO_AF_LOG_ADDR = string.Template(
        'Term $term will not be rendered, as it has'
        ' $direction address match specified but no'
        ' $direction addresses of $af address family'
        ' are present.'
    )

    NO_AF_LOG_PROTO = string.Template(
        'Term $term will not be rendered, as it has'
        ' $proto match specified but the ACL is of'
        ' $af address family.'
    )

    def __init__(self, term: policy.Term) -> None:
        if term.protocol:
            for protocol in term.protocol:
                if protocol not in self.PROTO_MAP and str(protocol) not in [
                    str(p) for p in self.PROTO_MAP_BY_NUMBER
                ]:
                    raise UnsupportedFilterError(
                        f'Protocol(s) {term.protocol!s} are not supported.'
                    )

            term.protocol = ProtocolNameToNumber(
                term.protocol, self.ALWAYS_PROTO_NUM, self.PROTO_MAP
            )
        self.term = term

    def NormalizeAddressFamily(self, af: Union[int, str]) -> int:
        """Convert (if necessary) address family name to numeric value.

        Args:
          af: Address family, can be either numeric or string (e.g. 4 or 'inet')

        Returns:
          af: Numeric address family value

        Raises:
          UnsupportedAFError: Address family not in keys or values of our AF_MAP.
        """
        # ensure address family (af) is valid
        if af in self.AF_MAP_BY_NUMBER:
            return af
        elif af in self.AF_MAP:
            # convert AF name to number (e.g. 'inet' becomes 4, 'inet6' becomes 6)
            af = self.AF_MAP[af]
        else:
            raise UnsupportedAFError(
                f'Address family {af} is not supported, term {self.term.name}.'
            )
        return af

    def NormalizeIcmpTypes(
        self, icmp_types: list[str], protocols: list[str], af: int
    ) -> list[int]:
        """Return verified list of appropriate icmp-types.

        Args:
          icmp_types: list of icmp_types
          protocols: list of protocols
          af: address family of this term, either numeric or text (see self.AF_MAP)

        Returns:
          sorted list of numeric icmp-type codes.

        Raises:
          UnsupportedFilterError: icmp-types specified with non-icmp protocol.
          MismatchIcmpInetError: mismatch between icmp protocol and address family.
          UnknownIcmpTypeError: unknown icmp-type specified
        """
        if not icmp_types:
            return ['']
        # only protocols icmp or icmpv6 can be used with icmp-types
        if (
            protocols != ['icmp']
            and protocols != ['icmpv6']
            and protocols != [self.PROTO_MAP['icmp']]
            and protocols != [self.PROTO_MAP['icmpv6']]
        ):
            raise UnsupportedFilterError(
                '{} {}'.format(
                    'icmp-types specified for non-icmp protocols in term: ', self.term.name
                )
            )
        # make sure we have a numeric address family (4 or 6)
        af = self.NormalizeAddressFamily(af)
        # check that addr family and protocl are appropriate
        if (af != 4 and protocols == ['icmp']) or (af != 6 and protocols == ['icmpv6']):
            raise MismatchIcmpInetError(
                '%s %s, %s: %s, %s: %s'
                % (
                    'ICMP/ICMPv6 mismatch with address family IPv4/IPv6 in term',
                    self.term.name,
                    'address family',
                    af,
                    'protocols',
                    ','.join(protocols),
                )
            )
        # ensure all icmp types are valid
        for icmptype in icmp_types:
            if icmptype not in self.ICMP_TYPE[af]:
                raise UnknownIcmpTypeError(
                    '%s %s %s %s'
                    % (
                        '\nUnrecognized ICMP-type (',
                        icmptype,
                        ') specified in term ',
                        self.term.name,
                    )
                )
        rval = []
        rval.extend([self.ICMP_TYPE[af][x] for x in icmp_types])
        rval.sort()
        return rval


class ACLGenerator:
    """Generates platform specific filters and terms from a policy object.

    This class takes a policy object and renders the output into a syntax which
    is understood by a specific platform (eg. iptables, cisco, etc).
    """

    SUFFIX: str = ''
    # Subclasses must override this value.

    _PLATFORM = None
    # Default protocol to apply when no protocol is specified.
    _DEFAULT_PROTOCOL = 'ip'
    # Unsupported protocols by address family.
    _SUPPORTED_AF = {'inet', 'inet6'}
    # Commonly misspelled protocols that the generator should reject.
    _FILTER_BLACKLIST = {}

    # Only warn if these tokens are not implemented by a platform. These are not
    # meant to be overridden in subclasses like supported tokens/sub tokens.
    WARN_IF_UNSUPPORTED = {
        'restrict_address_family',
        'counter',
        'destination_tag',
        'filter_term',
        'logging',
        'loss_priority',
        'owner',
        'qos',
        'routing_instance',
        'policer',
        'source_tag',
    }

    # Abbreviation table used to automatically abbreviate terms that exceed
    # specified limit. We use uppercase for abbreviations to distinguish
    # from lowercase names.  This is order list - we try the ones in the
    # top of the list before the ones later in the list.  Prefer clear
    # or very-space-saving abbreviations by putting them early in the
    # list.  Abbreviations may be regular expressions or fixed terms;
    # prefer fixed terms unless there's a clear benefit to regular
    # expressions.
    _ABBREVIATION_TABLE = [
        # Service abbreviations first.
        ('experiment', 'EXP'),
        ('wifi-radius', 'W-R'),
        ('customer', 'CUST'),
        ('server', 'SRV'),
        # Next, common routing terms
        ('global', 'GBL'),
        ('google', 'GOOG'),
        ('service', 'SVC'),
        ('router', 'RTR'),
        ('internal', 'INT'),
        ('external', 'EXT'),
        ('transit', 'TRNS'),
        ('management', 'MGMT'),
        # State info
        ('established', 'EST'),
        ('unreachable', 'UNR'),
        ('fragment', 'FRAG'),
        ('accept', 'ACC'),
        ('discard', 'DISC'),
        ('reject', 'REJ'),
        ('replies', 'RPL'),
        ('request', 'REQ'),
    ]
    # Maximum term length. Can be overridden by generator to enforce
    # platform specific restrictions.
    _TERM_MAX_LENGTH = 62

    def __init__(self, pol: policy.Policy, exp_info: int) -> None:
        """Initialise an ACLGenerator.  Store policy structure for processing."""
        supported_tokens, supported_sub_tokens = self._GetSupportedTokens()

        pol.filters[:] = [
            (header, terms)
            for (header, terms) in pol.filters
            if (self._PLATFORM in header.platforms)
        ]
        self.policy = pol
        all_err = []
        all_warn = []
        for header, terms in pol.filters:
            terms[:] = self._FilteredTerms(header, terms, exp_info)

            # Verify valid keywords
            # error on unsupported optional keywords that could result
            # in dangerous or unexpected results
            for term in terms:
                # Only verify optional keywords if the term is active on the platform.
                err = []
                warn = []
                for el, val in term.__dict__.items():
                    # Private attributes do not need to be valid keywords.
                    if val and el not in supported_tokens and not el.startswith('flatten'):
                        if val and el not in self.WARN_IF_UNSUPPORTED:
                            err.append(el)
                        else:
                            warn.append(el)
                    # ignore Liskov's rule.
                    if val and isinstance(val, list) and el in supported_sub_tokens:
                        ns = set(val) - supported_sub_tokens[el]
                        # hack support for ArbitraryOptions in junos. todo, add the
                        # junos options into the lexer, then we can nuke .*
                        # shenanigans.
                        if ns and '.*' not in supported_sub_tokens[el]:
                            err.append(' '.join(ns))
                if err:
                    all_err.append(
                        ('%s contains unsupported keywords (%s) for target %s in policy %s')
                        % (term.name, ' '.join(err), self._PLATFORM, pol.filename)
                    )
                if warn:
                    all_warn.append(
                        ('%s contains unimplemented keywords (%s) for target %s in policy %s')
                        % (term.name, ' '.join(warn), self._PLATFORM, pol.filename)
                    )
            continue
        if all_err:
            raise UnsupportedFilterError('\n %s' % '\n'.join(all_err))
        if all_warn:
            logging.debug('\n %s', '\n'.join(all_warn))
        self._TranslatePolicy(pol, exp_info)

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        # pylint: disable=unused-argument
        """Translate policy contents to platform specific data structures."""
        raise Error(f'{self._PLATFORM} does not implement _TranslatePolicies()')

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Provide a default for supported tokens and sub tokens.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        # Set of supported keywords for a given platform.  Values should be in
        # undercase form, eg, icmp_type (not icmp-type)
        supported_tokens = {
            'action',
            'comment',
            'destination_address',
            'destination_address_exclude',
            'destination_port',
            'expiration',
            'icmp_type',
            'stateless_reply',
            'name',  # obj attribute, not token
            'option',
            'protocol',
            'platform',
            'platform_exclude',
            'source_address',
            'source_address_exclude',
            'source_port',
            'translated',  # obj attribute, not token
            'verbatim',
        }

        # These keys must be also listed in supported_tokens.
        # Keys should be in undercase form, eg, icmp_type (not icmp-type). Values
        # should be in dash form, icmp-type (not icmp_type)
        supported_sub_tokens = {
            'option': {
                'established',
                'first-fragment',
                'is-fragment',
                'initial',
                'rst',
                'sample',
                'tcp-established',
            },
            'action': {
                'accept',
                'deny',
                'next',
                'reject',
                'reject-with-tcp-rst',
            },
            'icmp_type': set(list(Term.ICMP_TYPE[4].keys()) + list(Term.ICMP_TYPE[6].keys())),
        }
        return supported_tokens, supported_sub_tokens

    def _GetSupportedTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build our supported tokens and sub tokens.

        Returns:
          tuple containing the supported tokens and sub tokens.
        Raises:
          UnsupportedFilterError: Raised when token is not supported.
        """
        supported_tokens, supported_sub_tokens = self._BuildTokens()
        # make sure we don't have subtokens that are not listed. This should not
        # occur unless a platform's tokens/subtokens are changed.
        undefined_st = set(supported_sub_tokens) - supported_tokens
        if undefined_st:
            raise UnsupportedFilterError(
                'Found undefined sub tokens missing from the supported token list! '
                'These must match. (%s)' % ' '.join(undefined_st)
            )
        # all good.
        return supported_tokens, supported_sub_tokens

    # TODO(robankeny) Fix this function, it no longer does what it says.
    def FixHighPorts(
        self, term: policy.Term, af: str = 'inet', all_protocols_stateful: bool = False
    ):
        """Evaluate protocol and ports of term, return sane version of term.

        Args:
          term: Term object to be checked
          af: String presenting the address family, inet, inet6
          all_protocols_stateful: Boolean suggesting if protocols are all stateful.

        Returns:
          Copy of term that has been fixed

        Raises:
          UnsupportedAFError: Address family provided but unsupported.
          UnsupportedFilter: Protocols do not match the address family.
          EstablishedError: Established option used with inappropriate protocol.
          UnsupportedFilterError: Filter does not support protocols with AF.
        """
        mod = term

        # Determine which protocols this term applies to.
        if term.protocol:
            protocols = set(term.protocol)
        else:
            protocols = {self._DEFAULT_PROTOCOL}

        # Check that the address family matches the protocols.
        if af not in self._SUPPORTED_AF:
            raise UnsupportedAFError(
                '\nAddress family %s, found in %s, unsupported '
                'by %s' % (af, term.name, self._PLATFORM)
            )
        if af in self._FILTER_BLACKLIST:
            unsupported_protocols = self._FILTER_BLACKLIST[af].intersection(protocols)
            if unsupported_protocols:
                raise UnsupportedFilterError(
                    '\n%s targets do not support protocol(s) %s '
                    'with address family %s (in %s)'
                    % (self._PLATFORM, unsupported_protocols, af, term.name)
                )

        # Many renders expect high ports for terms with the established option.
        for opt in [str(x) for x in term.option]:
            if opt.find('established') == 0:
                unstateful_protocols = protocols.difference({'tcp', 'udp'})
                if not unstateful_protocols:
                    # TCP/UDP: add in high ports then collapse to eliminate overlaps.
                    mod = copy.deepcopy(term)
                    if not all_protocols_stateful:
                        mod.destination_port.append((1024, 65535))
                    mod.destination_port = mod.CollapsePortList(mod.destination_port)
                elif not all_protocols_stateful:
                    errmsg = 'Established option supplied with inappropriate protocol(s)'
                    raise EstablishedError(f'{errmsg} {unstateful_protocols} in term {term.name}')
                break

        return mod

    def FixTermLength(
        self,
        term_name: str,
        abbreviate: bool = False,
        truncate: bool = False,
        override_max_length: Optional[int] = None,
    ):
        """Return a term name which is equal or shorter than _TERM_MAX_LENGTH.

           New term is obtained in two steps. First, if allowed, automatic
           abbreviation is performed using hardcoded abbreviation table. Second,
           if allowed, term name is truncated to specified limit.

        Args:
          term_name: Name to abbreviate if necessary.
          abbreviate: Whether to allow abbreviations to shorten the length.
          truncate: Whether to allow truncation to shorten the length.
          override_max_length: Override the _TERM_MAX_LENGTH to a different value.
        Returns:
           A string based on term_name, that is equal or shorter than
           _TERM_MAX_LENGTH abbreviated and truncated as necessary.
        Raises:
           TermNameTooLongError: term_name cannot be abbreviated
           to be shorter than _TERM_MAX_LENGTH, or truncation is disabled.
        """
        new_term = term_name
        if override_max_length is None:
            override_max_length = self._TERM_MAX_LENGTH
        if abbreviate:
            for word, abbrev in self._ABBREVIATION_TABLE:
                if len(new_term) <= override_max_length:
                    return new_term
                new_term = re.sub(word, abbrev, new_term)
        if truncate:
            new_term = new_term[:override_max_length]
        if len(new_term) <= override_max_length:
            return new_term
        raise TermNameTooLongError(
            'Term %s (originally %s) is '
            'too long. Limit is %d characters (vs. %d) '
            'and no abbreviations remain or abbreviations '
            'disabled.' % (new_term, term_name, override_max_length, len(new_term))
        )

    def HexDigest(self, name: str, truncation_length: Optional[int] = None):
        """Return a hexadecimal digest of the name object.

        Args:
          name: Name to hash.
          truncation_length: Truncation to shorten the digest length if necessary.
        Returns:
          A hexadecimal digest of the input name.
        Raises:
          N/A
        """

        if truncation_length is None:
            truncation_length = 64
        name_bytes = name.encode('UTF-8')
        return hashlib.sha256(name_bytes).hexdigest()[:truncation_length]

    def _FilteredTerms(self, header: policy.Header, terms: list[policy.Term], exp_info: int):
        new_terms = []
        filter_name = header.FilterName(self._PLATFORM)
        current_date = datetime.datetime.now(datetime.timezone.utc).date()
        exp_info_date = current_date + datetime.timedelta(weeks=exp_info)

        for term in terms:
            if term.expiration:
                if term.expiration <= current_date:
                    logging.warning(
                        'WARNING: Term %s in policy %s is expired and will not be rendered.',
                        term.name,
                        filter_name,
                    )
                    continue
                elif term.expiration <= exp_info_date:
                    logging.info(
                        "INFO: Term %s in policy %s expires in less than two weeks.",
                        term.name,
                        filter_name,
                    )

            if (term.platform and self._PLATFORM not in term.platform) or (
                term.platform_exclude and self._PLATFORM in term.platform_exclude
            ):
                continue

            new_terms.append(term)

        return new_terms


def ProtocolNameToNumber(
    protocols: list[str], proto_to_num: list[str], name_to_num_map: dict[str, int]
) -> list[Union[str, int]]:
    """Convert a protocol name to a numeric value.

    Args:
      protocols: list of protocol names to inspect
      proto_to_num: list of protocol names that should be converted to numbers
      name_to_num_map: map of protocol names to protocol numbers

    Returns:
      return_proto: list of protocol names, converted if applicable
    """
    return_proto = []

    for protocol in protocols:
        if protocol in proto_to_num:
            return_proto.append(name_to_num_map[protocol])
        else:
            return_proto.append(protocol)

    return return_proto


def AddRepositoryTags(
    prefix: str = '',
    rid: bool = True,
    date: bool = True,
    revision: bool = True,
    wrap: bool = False,
) -> list[str]:
    """Add repository tagging into the output.

    Args:
      prefix: comment delimiter, if needed, to appear before tags
      rid: bool; True includes the revision Id: repository tag.
      date: bool; True includes the Date: repository tag.
      revision: bool; True includes the Revision: repository tag.
      wrap: bool; True wraps the tag in double quotes.
    Returns:
      list of text lines containing revision data
    """
    tags = []
    wrapper = '"' if wrap else ''

    p4_id = f'{wrapper}$Id:${wrapper}'
    p4_date = f'{wrapper}$Date:${wrapper}'
    p4_revision = f'{wrapper}$Revision:${wrapper}'
    if rid:
        tags.append(f'{prefix}{p4_id}')
    if date:
        tags.append(f'{prefix}{p4_date}')
    if revision:
        tags.append(f'{prefix}{p4_revision}')
    return tags


def WrapWords(textlist: list[str], size: int, joiner: str = '\n'):
    r"""Insert breaks into the listed strings at specified width.

    Args:
      textlist: a list of text strings
      size: width of reformated strings
      joiner: text to insert at break.  eg. '\n  ' to add an indent.
    Returns:
      list of strings
    """
    # \S{%d}(?!\s|\Z) collets the max size for words that are larger than the max
    # (?<=\S{%d})\S+ collects the remaining text for overflow words in their own
    # line
    # \S.{1,%d}(?=\s|\Z)) collects all words and spaces up to max size, breaking
    # at the last space
    rval = []
    linelength_re = re.compile(
        r'(\S{%d}(?!\s|\Z)|(?<=\S{%d})\S+|\S.{1,%d}(?=\s|\Z))' % (size, size, size - 1)
    )
    for index in range(len(textlist)):
        if len(textlist[index]) > size:
            # insert joiner into the string at appropriate places.
            textlist[index] = joiner.join(linelength_re.findall(textlist[index]))
        # avoid empty comment lines
        rval.extend(x.strip() for x in textlist[index].strip().split(joiner) if x)
    return rval



# Copyright 2016 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Windows IP security policy generator."""


# pylint: disable=g-importing-member
from string import Template
from typing import Union

from absl import logging

from aerleon.lib import aclgenerator, windows
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Header


class Term(windows.Term):
    """Generate windows IP security policy terms."""

    _PLATFORM = 'windows_ipsec'
    CMD_PREFIX = 'netsh ipsec static add '

    # Windows IPSec Policy (which actually isn't limit to IPSec proper, as you
    # might expect) is structured such that you create:
    #   * One policy (more can be defined, but only one active)
    #   * One or more filter lists, which is similar to an IP chain (but does not
    #     support action:: next)
    #   * A filter, which is the matcher portion of a term
    #   * A filteraction, which is the action to perform when a a filter is
    #     matched.  Not that the filter action does not support logging, logging
    #     is associated with a auditpol filterlist.
    #   * a rule, which links a policy, filter list, and filter action.

    # Logging:
    # auditpol /set /subcategory:"Filtering Platform Packet Drop" /success:enable
    #   /failure:enable
    # auditpol /set /subcategory:"IPsec Driver" /success:enable /failure:enable
    # auditpol /set /subcategory:"IPsec Main Mode" /success:enable /failure:enable
    # auditpol /set /subcategory:"IPsec Quick Mode" /success:enable
    #   /failure:enable
    # auditpol /set /subcategory:"IPsec Extended Mode" /success:enable
    #   /failure:enable

    _MASK_ATOM = Template('${dir}mask=${mask}')
    _ADDR_ATOM = Template('${dir}addr=${addr} ${mask}')
    _PROTO_ATOM = Template('protocol=${protocol}')
    _PORT_ATOM = Template('${dir}port=${port}')

    # ipsec is not stateful, we need mirrored=yes to get responses:
    _FILTER_FORMAT = Template('filter filterlist=${name} mirrored=yes ' '${atoms}')
    _FILTERACTION_FORMAT = Template('filteraction name=${name} ' 'action=${action}')
    _FILTERLIST_FORMAT = Template('filterlist name=$name ')
    _RULE_FORMAT = Template(
        'rule name=$name policy=$policy ' 'filterlist=$filterlist ' 'filteraction=$filteraction '
    )
    _COMMENT_FORMAT = Template(': $comment')
    _LIST_SUFFIX = '-list'
    _ACTION_SUFFIX = '-action'

    # filter rules
    _ACTION_TABLE = {
        'accept': 'permit',
        'deny': 'block',
        'reject': 'block',
    }

    def _HandleIcmpTypes(
        self, icmp_types: list[str], protocols: list[str]
    ) -> tuple[list[str], list[str]]:
        if icmp_types:
            raise aclgenerator.UnsupportedFilterError(
                '\n%s %s %s %s'
                % ('icmp types unsupported by', self._PLATFORM, '\nError in term:', self.term.name)
            )
        return ([''], protocols)

    def _HandlePorts(
        self, src_ports: list[tuple[int, int]], dst_ports: list[tuple[int, int]]
    ) -> tuple[list[str], list[str]]:
        # ports = Map the ports in a straight list since multiports aren't supported
        return (self._CollapsePortTuples(src_ports), self._CollapsePortTuples(dst_ports))

    def _HandlePreRule(self, ret_str: list[str]) -> None:
        ret_str.append(self._ComposeFilterList())
        ret_str.append(self._ComposeFilterAction(self._ACTION_TABLE[self.term.action[0]]))

    def _CartesianProduct(
        self,
        src_addr: list[Union[IPv4, IPv6]],
        dst_addr: list[Union[IPv4, IPv6]],
        protocol: list[str],
        unused_icmp_types: list[str],
        src_port: list[str],
        dst_port: list[str],
        ret_str: list[str],
    ) -> None:
        for saddr in src_addr:
            if saddr.version != 4:
                logging.warning(
                    'WARNING: term contains a non IPv4 address %s, '
                    'ignoring element of term %s.',
                    saddr,
                    self.term_name,
                )
                continue

            for daddr in dst_addr:
                if daddr.version != 4:
                    logging.warning(
                        'WARNING: term contains a non IPv4 address %s, '
                        'ignoring element of term %s.',
                        daddr,
                        self.term_name,
                    )
                    continue

                for proto in protocol:
                    for sport in src_port:
                        for dport in dst_port:
                            ret_str.append(
                                self._ComposeFilter(
                                    saddr.network_address,
                                    daddr.network_address,
                                    proto,
                                    saddr.prefixlen,
                                    daddr.prefixlen,
                                    sport,
                                    dport,
                                )
                            )

    def _CollapsePortTuples(self, port_tuples: tuple[int, int]) -> list[Union[str, int]]:
        ports = ['']
        for tpl in port_tuples:
            if tpl:
                (port_start, port_end) = tpl
                ports = list(range(port_start, port_end + 1))
        return ports

    def _ComposeFilterList(self) -> str:
        return self.CMD_PREFIX + self._FILTERLIST_FORMAT.substitute(
            name=self.term_name + self._LIST_SUFFIX
        )

    def _ComposeFilterAction(self, action) -> str:
        return self.CMD_PREFIX + self._FILTERACTION_FORMAT.substitute(
            name=self.term_name + self._ACTION_SUFFIX, action=action
        )

    def _ComposeFilter(
        self,
        srcaddr: Union[IPv4, IPv6],
        dstaddr: Union[IPv4, IPv6],
        proto: str,
        srcmask: int,
        dstmask: int,
        srcport: str,
        dstport: str,
    ):
        """Convert the given parameters to a netsh filter rule string."""
        atoms = []

        if srcmask == 0:
            atoms.append(self._ADDR_ATOM.substitute(dir='src', addr='any', mask=''))
        else:
            mask_atom = self._MASK_ATOM.substitute(dir='src', mask=srcmask)
            atoms.append(self._ADDR_ATOM.substitute(dir='src', addr=srcaddr, mask=mask_atom))

        if dstmask == 0:
            atoms.append(self._ADDR_ATOM.substitute(dir='dst', addr='any', mask=''))
        else:
            mask_atom = self._MASK_ATOM.substitute(dir='dst', mask=dstmask)
            atoms.append(self._ADDR_ATOM.substitute(dir='dst', addr=dstaddr, mask=mask_atom))

        if srcport:
            atoms.append(self._PORT_ATOM.substitute(dir='src', port=srcport))
        if dstport:
            atoms.append(self._PORT_ATOM.substitute(dir='dst', port=dstport))

        if proto:
            atoms.append(self._PROTO_ATOM.substitute(protocol=proto))

        return self.CMD_PREFIX + self._FILTER_FORMAT.substitute(
            name=self.term_name + self._LIST_SUFFIX, atoms=' '.join(atoms)
        )

    def ComposeRule(self, policy: str):
        return self.CMD_PREFIX + self._RULE_FORMAT.substitute(
            name=f"{self.term_name}-rule",
            policy=policy,
            filterlist=self.term_name + self._LIST_SUFFIX,
            filteraction=self.term_name + self._ACTION_SUFFIX,
        )


class WindowsIPSec(windows.WindowsGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'windows_ipsec'
    _TERM = Term

    _POLICY_FORMAT = Template('policy name=$name assign=yes')
    _POLICY_SUFFIX = '-policy'

    _GOOD_AFS = ['inet']

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens -= {'icmp_type'}
        del supported_sub_tokens['icmp_type']
        return supported_tokens, supported_sub_tokens

    def _HandlePolicyHeader(self, header: Header, target: list[str]) -> None:
        policy_name = header.FilterName(self._PLATFORM) + self._POLICY_SUFFIX
        target.append(f"{Term.CMD_PREFIX}{self._POLICY_FORMAT.substitute(name=policy_name)}\n")

    def _HandleTermFooter(self, header: Header, term: Term, target: list[str]):
        target.append(f"{term.ComposeRule(header.FilterName(self._PLATFORM))}\n")



import math
from collections.abc import MutableMapping
from typing import Optional, Union

from aerleon.lib import aclgenerator, policy
from aerleon.lib.nacaddr import ExcludeAddrs, IPv4, IPv6
from aerleon.lib.policy import PROTOS_WITH_PORTS, Policy
from aerleon.utils.options import BooleanKeywordOption as _BooleanKeywordOption
from aerleon.utils.options import (
    MultiValueOption,
    NumberValueOption,
    ProcessOptions,
    ValueOption,
)

### constants ###
LOG_LEVELS_MAP_OPTIONS = {
    'log_nolog': 'nolog',
    'log_emergency': 'emerg',
    'log_alert': 'alert',
    'log_critical': 'crit',
    'log_error': 'err',
    'log_warning': 'warning',
    'log_notice': 'notice',
    'log_info': 'info',
    'log_debug': 'debug',
}

ACTIONS_MAP = {
    'accept': 'ACCEPT',
    'deny': 'DROP',
    'reject': 'REJECT',
}


### error classes ###
class Error(aclgenerator.Error):
    pass


class UnsupportedFilterOptionError(Error):
    pass


class ZoneMismatchError(Error):
    pass


### helper classes ###
class ProxmoxConfigDataClass(MutableMapping):
    def __init__(self, *args, **kwargs):
        self._store: dict[str, Union[str, list[str]]] = dict()
        self._store.update(*args)
        self._store.update(**kwargs)
        self._store['enable'] = '1'

    def __len__(self):
        return len(self._store)

    def __iter__(self):
        return iter(self._store)

    def __getitem__(self, item):
        return self._store[item]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __add__(self, other):
        return self.merge(other)

    def keys(self):
        return self._store.keys()

    def flatten(self, item):
        if isinstance(item, dict):
            return {k: self.flatten(v) for k, v in item.items()}
        elif isinstance(item, list) or isinstance(item, set):
            return ",".join([self.flatten(i) for i in item])
        elif not isinstance(item, str):
            return str(item)
        else:
            return item

    @staticmethod
    def _merge_list(list_a: list, list_b: list) -> list:
        return sorted(list(set(list_a + list_b)))

    def merge(self, other):
        self_keys = list(self.keys())
        other_keys = list(other.keys())
        ret = dict()
        for k in self_keys + other_keys:
            if k in self_keys and k in other_keys:
                if isinstance(self[k], list) and isinstance(other[k], list):
                    ret[k] = self._merge_list(self[k], other[k])
                else:
                    ret[k] = other[k]
            elif k in self_keys:
                ret[k] = self[k]
            elif k in other_keys:
                ret[k] = other[k]
        self._store = ret
        return self

    def __str__(self):
        flattened_store = self.flatten(self._store)
        lines = []
        for k, v in flattened_store.items():
            lines.append(f'{k}: {v}')
        return "\n".join(lines)


class ProxmoxPort:
    @staticmethod
    def _singlePortFmt(port: int) -> str:
        return str(port)

    @staticmethod
    def _portTupleFmt(port: tuple) -> str:
        if port[0] == port[1]:
            return str(port[0])
        else:
            return f"{port[0]}-{port[1]}"

    def __init__(self, port: Union[tuple[int, int], int]):
        self.str_representation = ''
        if isinstance(port, int):
            self.str_representation = self._singlePortFmt(port)
        elif isinstance(port, tuple):
            self.str_representation = self._portTupleFmt(port)

    def __str__(self):
        return self.str_representation


class ProxmoxIcmp:
    # proxmox firewall only supports a subset of ICMP code/types
    ICMPv4_MAP = {
        None: {None: 'any'},
        'echo-reply': {None: 'echo-reply'},
        'unreachable': {
            None: 'destination-unreachable',
            0: 'network-unreachable',
            1: 'host-unreachable',
            2: 'protocol-unreachable',
            3: 'port-unreachable',
            4: 'fragmentation-needed',
            5: 'source-route-failed',
            6: 'network-unknown',
            7: 'host-unknown',
            9: 'network-prohibited',
            10: 'host-prohibited',
            11: 'TOS-network-unreachable',
            12: 'TOS-host-unreachable',
            13: 'communication-prohibited',
            14: 'host-precedence-violation',
            15: 'precedence-cutoff',
        },
        'source-quench': {None: 'source-quench'},
        'redirect': {
            None: 'redirect',
            0: 'network-redirect',
            1: 'host-redirect',
            2: 'TOS-network-redirect',
            3: 'TOS-host-redirect',
        },
        'echo-request': {None: 'echo-request'},
        'router-advertisement': {None: 'router-advertisement'},
        'router-solicitation': {None: 'router-solicitation'},
        'time-exceeded': {
            None: 'time-exceeded',
            0: 'ttl-zero-during-transit',
            1: 'ttl-zero-during-reassembly',
        },
        'parameter-problem': {
            None: 'parameter-problem',
            0: 'ip-header-bad',
            1: 'required-option-missing',
        },
        'timestamp-request': {None: 'timestamp-request'},
        'timestamp-reply': {None: 'timestamp-reply'},
        'mask-request': {None: 'address-mask-request'},
        'mask-reply': {None: 'address-mask-reply'},
    }
    ICMPv6_MAP = {
        None: {None: 'any'},
        'destination-unreachable': {
            None: 'destination-unreachable',
            0: 'no-route',
            1: 'communication-prohibited',
            2: 'beyond-scope',
            3: 'address-unreachable',
            4: 'port-unreachable',
            5: 'failed-policy',
            6: 'reject-route',
        },
        'packet-too-big': {None: 'packet-too-big'},
        'time-exceeded': {
            None: 'time-exceeded',
            0: 'ttl-zero-during-transit',
            1: 'ttl-zero-during-reassembly',
        },
        'parameter-problem': {
            None: 'parameter-problem',
            0: 'bad-header',
            1: 'unknown-header-type',
            2: 'unknown-option',
        },
        'echo-request': {None: 'echo-request'},
        'echo-reply': {None: 'echo-reply'},
        'router-solicit': {None: 'router-solicitation'},
        'router-advertisement': {None: 'router-advertisement'},
        'neighbor-solicit': {None: 'neighbour-solicitation'},
        'neighbor-advertisement': {None: 'neighbour-advertisement'},
        'redirect-message': {None: 'redirect'},
    }
    ICMP_MAP = {
        'icmp': ICMPv4_MAP,
        'icmpv6': ICMPv6_MAP,
        'icmp6': ICMPv6_MAP,
    }
    ICMP_PROTOS = ['icmp', 'icmp6', 'icmpv6']

    def __init__(
        self,
        icmp_proto: str,
        icmp_type: Optional[str] = None,
        icmp_code: Optional[int] = None,
    ):
        self.icmp_proto = icmp_proto
        self.type = icmp_type
        self.code = icmp_code

    def __str__(self):
        return self.ICMP_MAP[self.icmp_proto][self.type][self.code]


def BooleanKeywordOption(*args, **kwargs):
    return _BooleanKeywordOption(*args, **kwargs).withTrueValue("1")


### implementation ###
class Term(aclgenerator.Term):
    """
    (internal) representation of an individual proxmox-firewall term
    takes generic Term for conversion to internal type as an argument
    direction is IN, OUT, FORWARD (regardless of zone type, zone
    constraints should be handled by the client class)
    """

    _LOG_LEVELS_MAP = LOG_LEVELS_MAP_OPTIONS | {
        'true': 'warning',
        'disable': 'nolog',
    }
    # Proxmox uses slightly different names for certain IP protocols
    # since it depends on Debian's netbase package for /etc/protocols
    # (other distributions use different /etc/protocols files)
    PROXMOX_PROTO_MAP = {
        'icmpv6': 'ipv6-icmp',
        'icmp6': 'ipv6-icmp',
        'ipip': 'ipencap',
        'fragment': 'ipv6-frag',
    }

    def __init__(self, term: policy.Term, direction: str):
        self.term = term
        self.direction = direction

    @staticmethod
    def has_mixed_af(addresses: list[Union[IPv4, IPv6]]):
        has_v4 = any(map(lambda a: isinstance(a, IPv4), addresses))
        has_v6 = any(map(lambda a: isinstance(a, IPv6), addresses))
        return has_v4 and has_v6

    @staticmethod
    def filter_for_af(af: Union[type[IPv4], type[IPv6]], addresses: list[Union[IPv4, IPv6]]):
        return list(filter(lambda a: isinstance(a, af), addresses))

    @staticmethod
    def _ComputeAddresses(addresses, exclude_addresses):
        addresses_with_exclude = addresses
        if exclude_addresses:
            addresses_with_exclude = ExcludeAddrs(addresses, exclude_addresses)
        return addresses_with_exclude

    def __str__(self) -> str:
        """returns the proxmox-firewall string representation of the term"""
        ret_str = []
        address_families = [IPv4, IPv6]
        all_addresses = self.term.source_address + self.term.destination_address
        if not self.has_mixed_af(all_addresses) and all_addresses:  # single-AF term
            address_families = [type(all_addresses[0])]
        elif not all_addresses:  # no source and no destination
            address_families = [IPv6]  # doesn't matter which one but there should be only one

        icmp_types = self.term.icmp_type if self.term.icmp_type else [None]
        icmp_codes = self.term.icmp_code if self.term.icmp_code else [None]
        term_protocol = self.term.protocol if self.term.protocol else [None]

        # proxmox firewall only supports one protocol per rule
        for protocol in term_protocol:
            # proxmox firewall does not support mixed AFs in filters
            for af in address_families:
                for icmp_type in icmp_types:
                    for icmp_code in icmp_codes:
                        source = self._ComputeAddresses(
                            self.filter_for_af(af, self.term.source_address),
                            self.term.source_address_exclude,
                        )
                        dest = self._ComputeAddresses(
                            self.filter_for_af(af, self.term.destination_address),
                            self.term.destination_address_exclude,
                        )
                        ret_str.append(
                            self._Format(
                                protocol,
                                self.direction,
                                ACTIONS_MAP[self.term.action[0]],
                                source,
                                dest,
                                icmp_code,
                                icmp_type,
                                self.term.source_interface,
                                self.term.source_port,
                                self.term.destination_port,
                                self.term.comment,
                                self.term.logging,
                                self.term.option,
                            )
                        )

        return '\n'.join(ret_str)

    def _Format(
        self,
        protocol: Optional[str],
        direction: str,
        action: str,
        source_addresses: list[Union[IPv4, IPv6]],
        destination_addresses: list[Union[IPv4, IPv6]],
        icmp_code: Optional[str],
        icmp_type: Optional[str],
        source_interface: str,
        source_ports: list[Union[tuple[int, int], int]],
        destination_ports: list[Union[tuple[int, int], int]],
        comment: list[str],
        logging: list[str],
        term_options: list[str],
    ):
        def to_network_addr(i: Union[IPv6, IPv4]):
            return str(i.with_prefixlen)

        options = [direction, action]

        if protocol:
            options.append(f"-proto {self.PROXMOX_PROTO_MAP.get(protocol, protocol)}")

        # proxmox firewall supports multiple sources/destinations per rule
        if source_interface:
            options.append(f"-iface {source_interface}")

        # we cannot use address tokens since IPsets are defined in separate files (at cluster
        # or at host level) and we're only outputting one file at a time, which may not be
        # cluster/host firewall configuration.
        if destination_addresses:
            options.append(f"-dest {','.join(map(to_network_addr, destination_addresses))}")

        if source_addresses:
            options.append(f"-source {','.join(map(to_network_addr, source_addresses))}")

        if source_ports and protocol in PROTOS_WITH_PORTS:
            options.append(f"-sport {','.join(map(lambda p: str(ProxmoxPort(p)), source_ports))}")

        if destination_ports and protocol in PROTOS_WITH_PORTS:
            options.append(
                f"-dport {','.join(map(lambda p: str(ProxmoxPort(p)), destination_ports))}"
            )

        if protocol in ProxmoxIcmp.ICMP_PROTOS:
            options.append(f"-icmp-type {ProxmoxIcmp(protocol, icmp_type, icmp_code)}")

        if logging:
            log = self._GetLoggingLevel(term_options)
            options.append(f"-log {self._LOG_LEVELS_MAP[log]}")

        if comment:
            options.append(f"# {' '.join(comment)}")

        return ' '.join(options)

    def _GetLoggingLevel(self, term_options) -> str:
        logging_key = 'true'
        log_option = next(
            map(
                lambda o: o if o in self._LOG_LEVELS_MAP.keys() else None,
                term_options or [None],
            )
        )
        return log_option or logging_key


class Proxmox(aclgenerator.ACLGenerator):
    """Proxmox firewall policy object"""

    # aerleon class props
    _PLATFORM = 'proxmox'
    SUFFIX = '.prxmxfw'
    _LOG_LEVELS = list(LOG_LEVELS_MAP_OPTIONS.values())
    # own class props
    _GOOD_DIRECTIONS = ['IN', 'OUT', 'FORWARD']
    _NF_CONNTRACK_HELPERS = [
        "amanda",
        "ftp",
        "irc",
        "netbios-ns",
        "pptp",
        "sane",
        "sip",
        "snmp",
        "tftp",
    ]
    _BY_ZONE = {
        "cluster": {
            "supported_directions": _GOOD_DIRECTIONS,
            "supported_options": lambda config: [
                BooleanKeywordOption('enable', config),
                BooleanKeywordOption('ebtables', config),
                ValueOption(config, policy_forward=list(ACTIONS_MAP.values())),
                ValueOption(config, policy_in=list(ACTIONS_MAP.values())),
                ValueOption(config, policy_out=list(ACTIONS_MAP.values())),
            ],
        },
        "host": {
            "supported_directions": _GOOD_DIRECTIONS,
            "supported_options": lambda config: [
                BooleanKeywordOption('enable', config),
                ValueOption(config, log_level_forward=Proxmox._LOG_LEVELS),
                ValueOption(config, log_level_in=Proxmox._LOG_LEVELS),
                ValueOption(config, log_level_out=Proxmox._LOG_LEVELS),
                BooleanKeywordOption('log_nf_conntrack', config),
                BooleanKeywordOption('ndp', config),
                BooleanKeywordOption('nf_conntrack_allow_invalid', config),
                MultiValueOption(config, nf_conntrack_helpers=Proxmox._NF_CONNTRACK_HELPERS),
                NumberValueOption('nf_conntrack_max', 32768, math.inf, config),
                NumberValueOption('nf_conntrack_tcp_timeout_established', 7875, math.inf, config),
                NumberValueOption('nf_conntrack_tcp_timeout_syn_recv', 30, 60, config),
                BooleanKeywordOption('nftables', config),
                BooleanKeywordOption('nosmurfs', config),
                BooleanKeywordOption('protection_synflood', config),
                NumberValueOption('protection_synflood_burst', 0, math.inf, config),
                NumberValueOption('protection_synflood_rate', 0, math.inf, config),
                ValueOption(config, smurf_log_level=Proxmox._LOG_LEVELS),
                ValueOption(config, tcp_flags_log_level=Proxmox._LOG_LEVELS),
                BooleanKeywordOption('tcp_flags', config),
            ],
        },
        "vm": {
            "supported_directions": ["IN", "OUT"],
            "supported_options": lambda config: [
                BooleanKeywordOption('dhcp', config),
                BooleanKeywordOption('enable', config),
                BooleanKeywordOption('ipfilter', config),
                ValueOption(config, log_level_in=Proxmox._LOG_LEVELS),
                ValueOption(config, log_level_out=Proxmox._LOG_LEVELS),
                BooleanKeywordOption('macfilter', config),
                BooleanKeywordOption('ndp', config),
                ValueOption(config, policy_in=list(ACTIONS_MAP.values())),
                ValueOption(config, policy_out=list(ACTIONS_MAP.values())),
                BooleanKeywordOption('radv', config),
            ],
        },
        "vnet": {
            "supported_directions": ["FORWARD"],
            "supported_options": lambda config: [
                BooleanKeywordOption('enable', config),
                ValueOption(config, log_level_forward=Proxmox._LOG_LEVELS),
                ValueOption(config, policy_forward=list(ACTIONS_MAP.values())),
            ],
        },
    }

    def __init__(self, pol: Policy, exp_info: int):
        self.proxmox_policies = []
        super().__init__(pol, exp_info)

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """returns the list of DSL + YAML supplementary tokens supported (proxmox-firewall specific)"""
        supported_tokens, supported_sub_tokens = super()._BuildTokens()
        supported_tokens |= {
            # proxmox firewall only supports setting the source interface
            'source_interface',
            # proxmox firewall supports icmp type + icmp code
            'icmp_code',
        }
        supported_sub_tokens.update({'option': set(LOG_LEVELS_MAP_OPTIONS.keys())})
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        filter_zones = []
        global_policy_config = ProxmoxConfigDataClass()
        for header, terms in pol.filters:
            self.filter_options = header.FilterOptions(self._PLATFORM)

            if len(self.filter_options) < 2:
                raise UnsupportedFilterOptionError(
                    "missing options, zone and direction are mandatory"
                )

            filter_zone = self.filter_options[0]
            filter_direction = self.filter_options[1]

            if filter_zone not in self._BY_ZONE.keys():
                raise UnsupportedFilterOptionError("unknown zone")

            if filter_zones and filter_zone not in filter_zones:
                raise ZoneMismatchError("cannot mix zone types in one policy")
            filter_zones.append(filter_zone)

            if filter_direction not in self._BY_ZONE[filter_zone]["supported_directions"]:
                raise UnsupportedFilterOptionError(
                    "direction "
                    + filter_direction
                    + " not supported for zone type "
                    + filter_zone,
                )

            global_policy_config += ProcessOptions(
                self._BY_ZONE[filter_zone]['supported_options'],
                self.filter_options[2:],
                ProxmoxConfigDataClass(),
            )  # merge, will stay set to the same ref

            new_terms = []
            for term in terms:
                new_terms.append(
                    Term(
                        term,
                        filter_direction,
                    )
                )
            self.proxmox_policies.append(
                (header, filter_zone, filter_direction, global_policy_config, new_terms)
            )

    def __str__(self):
        target: list[str] = []
        terms_str: list[str] = []
        global_policy_config = ''
        for header, zone, direction, filter_config, terms in self.proxmox_policies:
            global_policy_config = filter_config  # only one (merged) config for the whole zone
            for term in terms:
                term_str = str(term)
                if term_str:
                    terms_str.append(term_str)
        target.append('[OPTIONS]')
        target.append(str(global_policy_config))
        target.append('[RULES]')
        target.append('\n'.join(terms_str))
        return '\n'.join(target) + '\n'



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utility to provide exploration of policy definition files.

Allows read only access of policy definition files.  The library
creates a Policy object, which has filters containing terms.

This library does no expansion on the tokens directly, such as in policy.py.

TODO: This library is currently incomplete, and does not allow access to
      every argument of a policy term.
"""

from aerleon.lib import naming


class Error(Exception):
    """Generic error class."""


class FileOpenError(Error):
    """Trouble opening a file."""


class InvalidFilterError(Error):
    """Filter is invalid."""


class Filter:
    """Simple filter with a name a list of terms."""

    def __init__(self, filtername=''):
        self.name = filtername
        self.term = []

    def __str__(self):
        rval = []
        title = f'Filter: {self.name!s}'
        rval.append(f'\n{title}')
        rval.append('-' * len(title))
        for term in self.term:
            rval.append(str(term))
        return '\n\n'.join(rval)


class Term:
    """Simple term with a name a list of attributes."""

    def __init__(self, termname=''):
        self.name = termname
        self.source = []
        self.destination = []
        self.sport = []
        self.dport = []
        self.action = []
        self.option = []
        self.protocol = []

    def __str__(self):
        rval = []
        rval.append(f'  Term: {self.name}')
        rval.append(f"  Source-address:: {' '.join(self.source)}")
        rval.append(f"  Destination-address:: {' '.join(self.destination)}")
        rval.append(f"  Source-port:: {' '.join(self.sport)}")
        rval.append(f"  Destination-port:: {' '.join(self.dport)}")
        rval.append(f"  Protocol:: {' '.join(self.protocol)}")
        rval.append(f"  Option:: {' '.join(self.option)}")
        rval.append(f"  Action:: {' '.join(self.action)}")
        return '\n'.join(rval)


class Policy:
    """Holds basic attributes of an unexpanded policy definition file."""

    def __init__(self, filename, defs_data=None):
        """Build policy object and naming definitions from provided filenames.

        Args:
          filename: location of a .pol file
          defs_data: location of naming definitions directory, if any
        """
        self.defs = naming.Naming(defs_data)
        self.filter = []
        try:
            self.data = open(filename).readlines()
        except OSError as error_info:
            info = f"{filename!s} cannot be opened"
            raise FileOpenError(f'{info}\n{error_info}')

        indent = 0
        in_header = False
        in_term = False
        filt = Filter()
        term = Term()
        in_string = False

        for line in self.data:
            words = line.strip().split()
            quotes = len(line.split('"')) + 1
            if quotes % 2:  # are we in or out of double quotes
                in_string = not in_string  # flip status of quote status
            if not in_string:
                if '{' in words:
                    indent += 1
                if words:
                    if words[0] == 'header':
                        in_header = True
                    if words[0] == 'term':
                        in_term = True
                        term = Term(words[1])
                    if in_header and words[0] == 'target::':
                        if filt.name != words[2]:  # avoid empty dupe filters due to
                            filt = Filter(words[2])  # multiple target header lines
                    if in_term:
                        if words[0] == 'source-address::':
                            term.source.extend(words[1:])
                        if words[0] == 'destination-address::':
                            term.destination.extend(words[1:])
                        if words[0] == 'source-port::':
                            term.sport.extend(words[1:])
                        if words[0] == 'destination-port::':
                            term.dport.extend(words[1:])
                        if words[0] == 'action::':
                            term.action.extend(words[1:])
                        if words[0] == 'protocol::':
                            term.protocol.extend(words[1:])
                        if words[0] == 'option::':
                            term.option.extend(words[1:])

                if '}' in words:
                    indent -= 1
                    if in_header:
                        self.filter.append(filt)
                        in_header = False
                    if in_term:
                        filt.term.append(term)
                        in_term = False

    def __str__(self):
        return '\n'.join(str(i) for i in self.filter)

    def Matches(self, src=None, dst=None, dport=None, sport=None, filtername=None):
        """Return list of term names that match specific attributes.

        Args:
          src: source ip address '12.1.1.1'
          dst: destination ip address '10.1.1.1'
          dport: any port/protocol combo, such as '80/tcp' or '53/udp'
          sport: any port/protocol combo, such as '80/tcp' or '53/udp'
          filtername: a filter name or None to search all filters

        Returns:
          results: list of lists, each list is index to filter & term in the policy

        Raises:
          InvalidFilterError: Error if filter is invalid.

        Example:
          p=policyreader.Policy('policy_path', 'definitions_path')

          p.Matches(dst='209.85.216.5', dport='25/tcp')
          [[0, 26]]
          print p.filter[0].term[26].name

          for match in p.Matches(dst='209.85.216.5'):
            print p.filter[match[0]].term[match[1]].name

        """
        rval = []
        results = []
        filter_list = []
        dport_parents = None
        sport_parents = None
        destination_parents = None
        source_parents = None
        if dport:
            dport_parents = self.defs.GetServiceParents(dport)
        if sport:
            sport_parents = self.defs.GetServiceParents(sport)
        if dst:
            destination_parents = self.defs.GetIpParents(dst)
            try:
                destination_parents.remove('ANY')
                destination_parents.remove('RESERVED')
            except ValueError:
                pass  # ignore and continue
        if src:
            source_parents = self.defs.GetIpParents(src)
            try:
                source_parents.remove('ANY')
                source_parents.remove('RESERVED')
            except ValueError:
                pass  # ignore and continue
        if not filtername:
            filter_list = self.filter
        else:
            for idx, fil in enumerate(self.filter):
                if filtername == fil.name:
                    filter_list = [self.filter[idx]]
            if not filter_list:
                raise InvalidFilterError(f'invalid filter name: {filtername}')

        for findex, xfilter in enumerate(filter_list):
            mterms = []
            mterms.append(set())  # dport
            mterms.append(set())  # sport
            mterms.append(set())  # dst
            mterms.append(set())  # src
            for tindex, term in enumerate(xfilter.term):
                if dport_parents:
                    for token in dport_parents:
                        if token in term.dport:
                            mterms[0].add(tindex)
                else:
                    mterms[0].add(tindex)
                if sport_parents:
                    for token in sport_parents:
                        if token in term.sport:
                            mterms[1].add(tindex)
                else:
                    mterms[1].add(tindex)
                if destination_parents:
                    for token in destination_parents:
                        if token in term.destination:
                            mterms[2].add(tindex)
                else:
                    mterms[2].add(tindex)
                if source_parents:
                    for token in source_parents:
                        if token in term.source:
                            mterms[3].add(tindex)
                else:
                    mterms[3].add(tindex)
            rval.append(list(mterms[0] & mterms[1] & mterms[2] & mterms[3]))
        for findex, fresult in enumerate(rval):
            for i in list(fresult):
                results.append([findex, i])
        return results



# Copyright 2023 Nokia All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Nokia SR Linux yang ACL generator.

More information about the SR Linux ACL model schema: https://yang.srlinux.dev/

"""

from typing import TypedDict

from aerleon.lib import aclgenerator, openconfig
from aerleon.lib.policy import Term

R24_3_2 = "r24.3.2"  # Option flag to generate release >= 24.3.2 syntax


class IPPrefix(TypedDict):
    prefix: str


class PortRange(TypedDict):
    start: int
    end: int


class Port(TypedDict):
    value: int
    range: PortRange


Match = TypedDict(
    "Match",
    {
        "fragment": bool,
        "first-fragment": bool,
        "protocol": int,
        "next-header": int,
        "source-ip": IPPrefix,
        "destination-ip": IPPrefix,
        "source-port": Port,
        "destination-port": Port,
    },
)


class Action(TypedDict):
    accept: None
    drop: None


ACLEntry = TypedDict(
    "ACLEntry",
    {
        "sequence-id": int,
        "action": Action,
        "match": Match,
        "description": str,
        "_annotate_description": str,
    },
)
Entries = TypedDict(
    "Entries",
    {
        "entry": list[ACLEntry],
        "description": str,
        "name": str,
        "type": str,
        "_annotate": str,
        "statistics-per-entry": bool,
    },
)
IPFilters = TypedDict("IPFilters", {"acl-filter": Entries})


# generic error class
class Error(aclgenerator.Error):
    pass


class TcpEstablishedWithNonTcpError(Error):
    pass


class EstablishedWithNoProtocolError(Error):
    pass


class EstablishedWithNonTcpUdpError(Error):
    pass


class UnsupportedLogging(Error):
    pass


class SRLTerm(openconfig.Term):
    """Creates the term for the SR Linux ACL."""

    ACTION_MAP = {'accept': 'accept', 'deny': 'drop', 'reject': 'drop'}

    def SetName(self, name: str) -> None:
        # Put name in description field
        self.term_dict['description'] = name

    def SetAction(self, filter_options: list[str]) -> None:
        action = self.ACTION_MAP[self.term.action[0]]
        if R24_3_2 in filter_options:
            self.term_dict['action'] = {action: {}}
            if self.term.logging:
                self.term_dict['action']['log'] = True
        else:
            log = {}
            if self.term.logging:
                if action == 'drop':
                    log = {"log": True}
                else:
                    raise UnsupportedLogging(
                        f'logging can only be used with deny in term {self.term.name}'
                    )
            self.term_dict['action'] = {action: log}

    def SetComments(self, comments: list[str]) -> None:
        self.term_dict['_annotate_description'] = "_".join(comments)[:255]

    # Handles syntax changes in release 24.3.2 and beyond
    def _field(self, key, filter_options: list[str]):
        if R24_3_2 in filter_options:
            if key not in self.term_dict['match']:
                self.term_dict['match'][key] = {}
            return self.term_dict['match'][key]
        return self.term_dict['match']

    def SetOptions(self, family: str, filter_options: list[str]) -> None:
        # Handle various options
        opts = [str(x) for x in self.term.option]
        self.term_dict['match'] = {}
        if ('fragments' in opts) or ('is-fragment' in opts):
            self._field('ipv4', filter_options)['fragment'] = True
        if 'first-fragment' in opts:
            self._field('ipv4', filter_options)['first-fragment'] = True

        if 'initial' in opts or 'tcp-initial' in opts:
            self._field('transport', filter_options)['tcp-flags'] = "syn"
        if 'rst' in opts:
            _f = self._field('transport', filter_options)
            _f['tcp-flags'] = "syn&rst" if 'tcp-flags' in _f else "rst"
        if 'not-syn-ack' in opts:
            self._field('transport', filter_options)['tcp-flags'] = "!(syn&ack)"

        def _tcp_established():
            self._field('transport', filter_options)['tcp-flags'] = "ack|rst"

        if 'tcp-established' in opts:
            if not self.term.protocol or self.term.protocol == ['tcp']:
                _tcp_established()
            else:
                raise TcpEstablishedWithNonTcpError(
                    f'tcp-established can only be used with tcp protocol in term {self.term.name}'
                )
        elif 'established' in opts:
            if self.term.protocol:
                if self.term.protocol == ['tcp']:
                    _tcp_established()
                elif self.term.protocol == ['udp']:
                    self.SetProtocol(family=family, protocol="udp", filter_options=filter_options)
                    if not self.term.destination_port:
                        self.SetDestPorts(1024, 65535, filter_options)
                else:  # Could produce 2 rules if [tcp,udp]
                    raise EstablishedWithNonTcpUdpError(
                        f'established can only be used with tcp or udp protocol in term {self.term.name}'
                    )
            else:
                raise EstablishedWithNoProtocolError(
                    f'must specify a protocol for "established" in term {self.term.name}'
                )

        if 'tcp-flags' in self.term_dict['match'] or (
            R24_3_2 in filter_options
            and 'transport' in self.term_dict['match']
            and 'tcp-flags' in self.term_dict['match']['transport']
        ):
            self.SetProtocol(family=family, protocol="tcp", filter_options=filter_options)

    def SetSourceAddress(self, family: str, saddr: str, filter_options: list[str]) -> None:
        self._field(family, filter_options)['source-ip'] = {'prefix': saddr}

    def SetDestAddress(self, family: str, daddr: str, filter_options: list[str]) -> None:
        self._field(family, filter_options)['destination-ip'] = {'prefix': daddr}

    def SetSourcePorts(self, start: int, end: int, filter_options: list[str]) -> None:
        if start == end:
            val = {'value': start}
        else:
            val = {'range': {'start': start, 'end': end}}
        self._field('transport', filter_options)['source-port'] = val

    def SetDestPorts(self, start: int, end: int, filter_options: list[str]) -> None:
        if start == end:
            val = {'value': start}
        else:
            val = {'range': {'start': start, 'end': end}}
        self._field('transport', filter_options)['destination-port'] = val

    def SetProtocol(self, family: str, protocol: int, filter_options: list[str]) -> None:
        field_name = "protocol" if family == "ipv4" else "next-header"
        self._field(family, filter_options)[field_name] = protocol


class NokiaSRLinux(openconfig.OpenConfig):
    """A Nokia SR Linux ACL object, derived from OpenConfig."""

    _PLATFORM = 'nokiasrl'
    SUFFIX = '.srl_acl'

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens -= {'platform', 'platform_exclude', 'verbatim', 'icmp-type'}

        supported_sub_tokens['action'] = {'accept', 'deny'}  # excludes 'reject'
        supported_sub_tokens['option'] = {
            'established',
            'first-fragment',
            'is-fragment',
            'fragments',
            #  'sample',
            'tcp-established',
            'tcp-initial',
            #  'inactive',
            'not-syn-ack',
        }
        return supported_tokens, supported_sub_tokens

    def _InitACLSet(self) -> None:
        """Initialize self.acl_sets with proper Typing"""
        self.acl_sets: list[IPFilters] = []

    def _TranslateTerms(
        self,
        terms: list[Term],
        address_family: str,
        filter_name: str,
        hdr_comments: list[str],
        filter_options: list[str],
    ) -> None:
        srl_acl_entries: dict[str, list[ACLEntry]] = {'inet': [], 'inet6': []}
        afs = ['inet', 'inet6'] if address_family == 'mixed' else [address_family]
        for term in terms:
            for term_af in afs:
                t = SRLTerm(term, term_af)
                for rule in t.ConvertToDict(filter_options):
                    self.total_rule_count += 1
                    rule['sequence-id'] = (len(srl_acl_entries[term_af]) + 1) * 5
                    srl_acl_entries[term_af].append(rule)
        desc = "_".join(hdr_comments)[:255] if hdr_comments else ""

        for af in srl_acl_entries.keys():
            if srl_acl_entries[af]:
                # r24.3 changed the syntax for filters. For r24.3 or higher use the option `r24.3`.
                if 'r24.3' in filter_options or R24_3_2 in filter_options:
                    key = "acl-filter"
                else:
                    key = "ipv4-filter" if af == 'inet' else "ipv6-filter"

                ip_filter = {
                    key: {
                        '_annotate': " ".join(aclgenerator.AddRepositoryTags()),
                        'name': filter_name,
                        'description': desc,
                        'entry': srl_acl_entries[af],
                    }
                }
                if 'stats' in filter_options:
                    ip_filter[key]['statistics-per-entry'] = True
                if 'r24.3' in filter_options or R24_3_2 in filter_options:
                    ip_filter[key]['type'] = "ipv4" if af == 'inet' else "ipv6"
                self.acl_sets.append(ip_filter)



from yaml.loader import SafeLoader


def SpanSafeYamlLoader(*, filename):
    """Configure yaml.load to:
    * Force safe_load mode (disable unpickling).
    * Augment mappings with debug context: __line__, __filename__.

    Post-load user error messages need to provide a filename and line number back to the user.
    Including debugging context in the mappings gives post-load code access to this information.
    Code operating on the native representation must filter out __line__, __filename__ from all
    mappings (dicts) when iterating over user data. This assumes __line__, __filename__ are not
    valid keys in any user data.
    """

    class PluginYamlLoader(SafeLoader):
        def construct_mapping(self, node, deep=False):
            mapping = super().construct_mapping(node, deep=deep)
            # Add 1 so line numbering starts at 1
            # TODO(jb) look at cases where line number does not match up, e.g. filter['__line__']
            mapping['__line__'] = node.start_mark.line + 1
            mapping['__filename__'] = filename
            return mapping

    return PluginYamlLoader



# Copyright 2012 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""PacketFilter (PF) generator."""

import collections
import copy
from typing import Optional, Union, cast

from absl import logging

from aerleon.lib import aclgenerator, nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Policy, Term, VarType


class Error(aclgenerator.Error):
    """Base error class."""


class DuplicateTermError(Error):
    """Raised when duplication of term names are detected."""


class DuplicateShortenedTableNameError(Error):
    """Raised when a duplicate shortened table name is found."""


class UnsupportedProtoError(Error):
    """Raised when a protocol is not supported."""


class Term(aclgenerator.Term):
    """Generate PacketFilter policy terms."""

    # Validate that term does not contain any fields we do not
    # support.  This prevents us from thinking that our output is
    # correct in cases where we've omitted fields from term.
    _PLATFORM = 'packetfilter'
    _ACTION_TABLE = {
        'accept': 'pass',
        'deny': 'block drop',
        'reject': 'block return',
        'next': 'pass',
    }
    # Moving the log keyword into an member variable allows subclasses to override
    # it to support logging options outside of the scope of the Aerleon policy
    # spec, e.g., on platform-specific options such as packetfilter's
    # "log (all, to pflogN)" per-direction.
    _LOG_TABLE = {
        '': 'log',
        'in': 'log',
        'out': 'log',
    }
    _QUICK_TABLE = {
        'accept': 'quick',
        'deny': 'quick',
        'reject': 'quick',
        'next': '',
    }
    _DIRECTION_TABLE = {
        '': '',
        'in': 'in',
        'out': 'out',
    }
    _TCP_FLAGS_TABLE = {
        'syn': 'S',
        'ack': 'A',
        'fin': 'F',
        'rst': 'R',
        'urg': 'U',
        'psh': 'P',
        'all': 'SAFRUP',
        'none': 'NONE',
    }
    _PROTO_TABLE = {
        'icmpv6': 'ipv6-icmp',
    }
    _UNSUPPORTED_PROTOS = ['hopopt']

    def __init__(
        self,
        term: Term,
        filter_name: str,
        stateful: bool = True,
        af: str = 'inet',
        direction: str = '',
    ) -> None:
        """Setup a new term.

        Args:
          term: A policy.Term object to represent in packetfilter.
          filter_name: The name of the filter chan to attach the term to.
          stateful: Whether to keep firewall state for the term.
          af: Which address family ('inet' or 'inet6') to apply the term to.
          direction: What direction the term applies to ('in', 'out' or both).

        Raises:
          aclgenerator.UnsupportedFilterError: Filter is not supported.
        """
        super().__init__(term)
        self.term = term  # term object
        self.filter = filter_name  # actual name of filter
        self.options = []
        self.default_action = 'deny'
        self.af = af
        self.stateful = stateful
        self.direction = direction

    def __str__(self) -> str:
        """Render config output from this term object."""

        ret_str = []
        self._SetDefaultAction()

        # Create a new term
        ret_str.append(f'\n# term {self.term.name}')

        comments = aclgenerator.WrapWords(self.term.comment, 80)
        # append comments to output
        if comments and comments[0]:
            for line in comments:
                ret_str.append(f'# {line!s}')

        if str(self.term.action[0]) not in self._ACTION_TABLE:
            raise aclgenerator.UnsupportedFilterError(
                '%s %s %s %s'
                % ('\n', self.term.name, self.term.action[0], 'action not currently supported.')
            )

        if self.direction and str(self.direction) not in self._DIRECTION_TABLE:
            raise aclgenerator.UnsupportedFilterError(
                '%s %s %s %s'
                % ('\n', self.term.name, self.term.direction, 'direction not currently supported.')
            )
        # protocol
        if self.term.protocol:
            protocol = self.term.protocol
        else:
            protocol = []

        # source address
        term_saddrs = self._CheckAddressAf(self.term.source_address)
        if not term_saddrs:
            if self.af != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_ADDR.substitute(
                        term=self.term.name, direction='source', af=self.af
                    )
                )
            return ''
        term_saddr = self._GenerateAddrStatement(term_saddrs, self.term.source_address_exclude)

        # destination address
        term_daddrs = self._CheckAddressAf(self.term.destination_address)
        if not term_daddrs:
            if self.af != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_ADDR.substitute(
                        term=self.term.name, direction='destination', af=self.af
                    )
                )
            return ''
        term_daddr = self._GenerateAddrStatement(
            term_daddrs, self.term.destination_address_exclude
        )

        # ports
        source_port = []
        destination_port = []
        if self.term.source_port:
            source_port = self._GeneratePortStatement(self.term.source_port)
        if self.term.destination_port:
            destination_port = self._GeneratePortStatement(self.term.destination_port)

        # icmp-type
        icmp_types = ['']
        if self.term.icmp_type:
            if protocol == ['icmp']:
                af = 'inet'
            elif protocol == ['icmpv6']:
                af = 'inet6'
            else:
                raise aclgenerator.UnsupportedFilterError(
                    '%s %s %s'
                    % ('\n', self.term.name, 'icmp protocol is not defined or not supported.')
                )
            icmp_types = self.NormalizeIcmpTypes(self.term.icmp_type, protocol, af)

        # options
        tcp_flags_set = []
        tcp_flags_check = []
        for next_opt in [str(x) for x in self.term.option]:
            for next_flag in self._TCP_FLAGS_TABLE:
                if next_opt.find(next_flag) == 0:
                    if protocol != ['tcp']:
                        raise aclgenerator.UnsupportedFilterError(
                            '%s %s %s'
                            % (
                                '\n',
                                self.term.name,
                                'tcp flags may only be specified with tcp protocol.',
                            )
                        )
                    tcp_flags_set.append(self._TCP_FLAGS_TABLE.get(next_flag))
                    tcp_flags_check.append(self._TCP_FLAGS_TABLE.get(next_flag))

        # If tcp-established is set, override any of the flags above with the
        # S/SA flags.  Issue an error if flags are specified with 'established'.
        for opt in [str(x) for x in self.term.option]:
            if opt == 'established' or opt == 'tcp-established':
                if tcp_flags_set or tcp_flags_check:
                    raise aclgenerator.UnsupportedFilterError(
                        '%s %s %s'
                        % (
                            '\n',
                            self.term.name,
                            'tcp flags may not be specified with tcp-established.',
                        )
                    )
                # We need to set 'flags A/A' for established regardless of whether or
                # not we're stateful:
                # - if we stateful, the default is 'flags S/SA' which prevent writing
                # rules for reply packets.
                # - if we're stateless, this is the only way to do it.
                if not protocol or 'tcp' in protocol:
                    tcp_flags_set.append(self._TCP_FLAGS_TABLE.get('ack'))
                    tcp_flags_check.append(self._TCP_FLAGS_TABLE.get('ack'))

        # The default behavior of pf is 'keep state flags S/SA'.  If we're not
        # stateless, and if flags have not been specified explicitly via options,
        # append that here.  Note that pf allows appending flags for udp and icmp;
        # they are just ignored, as long as TCP is in the proto.  This lets you
        # doing things like 'proto { tcp udp icmp } flags S/SA' and have the flags
        # only applied to the tcp bits that match.  However, the policy description
        # language prohibits setting flags on non-TCP, since it doesn't make sense
        # on all platforms.
        if (
            (not protocol or protocol == ['tcp'])
            and self.stateful
            and not tcp_flags_set
            and not tcp_flags_check
        ):
            tcp_flags_set.append(self._TCP_FLAGS_TABLE.get('syn'))
            tcp_flags_check.append(self._TCP_FLAGS_TABLE.get('syn'))
            tcp_flags_check.append(self._TCP_FLAGS_TABLE.get('ack'))

        ret_str.extend(
            self._FormatPart(
                self.term.action[0],
                self.direction,
                self.term.logging,
                self.term.interface,
                self.af,
                protocol,
                term_saddr,
                source_port,
                term_daddr,
                destination_port,
                tcp_flags_set,
                tcp_flags_check,
                icmp_types,
                self.options,
                self.stateful,
            )
        )

        return '\n'.join(str(v) for v in ret_str if v)

    def _CheckAddressAf(self, addrs: list[Union[IPv4, IPv6]]) -> list[Union[IPv4, IPv6, str]]:
        """Verify that the requested address-family matches the address's family."""
        if not addrs:
            return ['any']
        if self.af == 'mixed':
            return addrs
        af_addrs = []
        af = self.NormalizeAddressFamily(self.af)
        for addr in addrs:
            if addr.version == af:
                af_addrs.append(addr)
        return af_addrs

    def _FormatPart(
        self,
        action: str,
        direction: str,
        log: list[VarType],
        interface: Optional[str],
        af: str,
        proto: list[str],
        src_addr: str,
        src_port: Union[list[str], str],
        dst_addr: str,
        dst_port: Union[list[str], str],
        tcp_flags_set: list[str],
        tcp_flags_check: list[str],
        icmp_types: list[Union[str, int]],
        options: list[str],
        stateful: bool,
    ) -> list[str]:
        """Format the string which will become a single PF entry."""
        line = [f'{self._ACTION_TABLE.get(action)}']

        if direction:
            line.append(direction)

        quick = self._QUICK_TABLE.get(action)
        if quick:
            line.append(f'{quick}')

        if log:
            logaction = self._LOG_TABLE.get(direction)
            if logaction:
                line.append(logaction)
            else:
                line.append('log')

        if interface:
            line.append(f'on {interface}')

        if af != 'mixed':
            line.append(af)

        if proto:
            line.append(self._GenerateProtoStatement(proto))

        line.append(f'from {src_addr}')
        if src_port:
            line.append(f'port {src_port}')

        line.append(f'to {dst_addr}')
        if dst_port:
            line.append(f'port {dst_port}')

        if tcp_flags_set and tcp_flags_check:
            line.append('flags')
            line.append(f"{''.join(tcp_flags_set)}/{''.join(tcp_flags_check)}")

        if 'icmp' in proto and icmp_types:
            type_strs = [str(icmp_type) for icmp_type in icmp_types]
            type_strs = ', '.join(type_strs)
            if type_strs:
                line.append('icmp-type { %s }' % type_strs)

        if 'icmpv6' in proto and icmp_types:
            type_strs = [str(icmp_type) for icmp_type in icmp_types]
            type_strs = ', '.join(type_strs)
            if type_strs:
                line.append('icmp6-type { %s }' % type_strs)

        if options:
            line.extend(options)

        if not stateful:
            line.append('no state')
        elif action in ['accept', 'next']:
            line.append('keep state')

        return [' '.join(line)]

    def _GenerateProtoStatement(self, protocols: list[str]) -> str:
        proto = ''
        if protocols:
            protocols = copy.deepcopy(protocols)
            for i, proto in enumerate(protocols):
                if proto in self._UNSUPPORTED_PROTOS:
                    raise UnsupportedProtoError
                try:
                    protocols[i] = self._PROTO_TABLE[proto]
                except KeyError:
                    pass
            proto = 'proto { %s }' % ' '.join(protocols)
        return proto

    def _GenerateAddrStatement(
        self, addrs: list[Union[IPv4, IPv6]], exclude_addrs: list[Union[IPv4, IPv6]]
    ) -> str:
        addresses = set()
        if addrs != ['any']:
            parent_token_set = set()
            for addr in addrs:
                addr = cast(nacaddr.IPType, addr)
                parent_token_set.add(addr.parent_token)
            for token in parent_token_set:
                addresses.add(f'<{token[:31]}>')
        else:
            addresses.add('any')
        if exclude_addrs != ['any']:
            parent_token_set = set()
            for addr in exclude_addrs:
                addr = cast(nacaddr.IPType, addr)
                parent_token_set.add(addr.parent_token)
            for token in parent_token_set:
                addresses.add(f'!<{token[:31]}>')
        return '{ %s }' % ', '.join(sorted(addresses))

    def _GeneratePortStatement(self, ports: list[tuple[int, int]]) -> str:
        port_list = []
        for port_tuple in ports:
            if port_tuple[0] == port_tuple[1]:
                port_list.append(str(port_tuple[0]))
            else:
                port_list.append(f'{port_tuple[0]}:{port_tuple[1]}')
        return '{ %s }' % (' '.join(list(collections.OrderedDict.fromkeys(port_list))))

    def _SetDefaultAction(self) -> None:
        """If term does not specify action, use filter default action."""
        if not self.term.action:
            self.term.action[0].value = self.default_action


class PacketFilter(aclgenerator.ACLGenerator):
    """Generates filters and terms from provided policy object."""

    _DEF_MAX_LENGTH = 31
    _PLATFORM = 'packetfilter'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = '.pf'
    _TERM = Term

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'logging', 'destination_interface', 'source_interface'}
        supported_sub_tokens.update(
            {
                'action': {'accept', 'deny', 'reject', 'next'},
                'option': {
                    'established',
                    'tcp-established',
                    'syn',
                    'ack',
                    'fin',
                    'rst',
                    'urg',
                    'psh',
                    'all',
                },
            }
        )

        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        self.pf_policies = []
        self.address_book = {}
        self.def_short_to_long = {}
        good_afs = ['inet', 'inet6', 'mixed']
        good_options = ['in', 'out', 'nostate']
        all_protocols_stateful = True

        for header, terms in pol.filters:
            filter_type = None
            filter_options = header.FilterOptions(self._PLATFORM)[1:]
            filter_name = header.FilterName(self._PLATFORM)
            direction = ''

            # ensure all options after the filter name are expected
            for opt in filter_options:
                if opt not in good_afs + good_options:
                    raise aclgenerator.UnsupportedTargetOptionError(
                        '%s %s %s %s'
                        % (
                            '\nUnsupported option found in',
                            self._PLATFORM,
                            'target definition:',
                            opt,
                        )
                    )

            # pf will automatically add 'keep state flags S/SA' to all TCP connections
            # by default.
            if 'nostate' in filter_options:
                all_protocols_stateful = False

            if 'in' in filter_options:
                direction = 'in'
            elif 'out' in filter_options:
                direction = 'out'

            # Check for matching af
            for address_family in good_afs:
                if address_family in filter_options:
                    # should not specify more than one AF in options
                    if filter_type is not None:
                        raise aclgenerator.UnsupportedFilterError(
                            '%s %s %s %s'
                            % (
                                '\nMay only specify one of',
                                good_afs,
                                'in filter options:',
                                filter_options,
                            )
                        )
                    filter_type = address_family
            if filter_type is None:
                filter_type = 'inet'

            # add the terms
            new_terms = []
            term_names = set()

            for term in terms:
                term.name = self.FixTermLength(term.name)
                if term.name in term_names:
                    raise DuplicateTermError(f'You have a duplicate term: {term.name}')
                term_names.add(term.name)

                for source_addr in term.source_address:
                    src_token = source_addr.parent_token[: self._DEF_MAX_LENGTH]

                    if (
                        src_token in self.def_short_to_long
                        and self.def_short_to_long[src_token] != source_addr.parent_token
                    ):
                        raise DuplicateShortenedTableNameError(
                            'There is a shortened name conflict between names %s and %s '
                            '(different named objects would conflict when shortened to %s)'
                            % (
                                self.def_short_to_long[src_token],
                                source_addr.parent_token,
                                src_token,
                            )
                        )
                    else:
                        self.def_short_to_long[src_token] = source_addr.parent_token

                    if src_token not in self.address_book:
                        self.address_book[src_token] = {source_addr}
                    else:
                        self.address_book[src_token].add(source_addr)

                for dest_addr in term.destination_address:
                    dst_token = dest_addr.parent_token[: self._DEF_MAX_LENGTH]

                    if (
                        dst_token in self.def_short_to_long
                        and self.def_short_to_long[dst_token] != dest_addr.parent_token
                    ):
                        raise DuplicateShortenedTableNameError(
                            'There is a shortened name conflict between names %s and %s '
                            '(different named objects would conflict when shortened to %s)'
                            % (
                                self.def_short_to_long[dst_token],
                                dest_addr.parent_token,
                                dst_token,
                            )
                        )
                    else:
                        self.def_short_to_long[dst_token] = dest_addr.parent_token

                    if dst_token not in self.address_book:
                        self.address_book[dst_token] = {dest_addr}
                    else:
                        self.address_book[dst_token].add(dest_addr)

                if not term:
                    continue

                if term.destination_interface and term.source_interface:
                    raise Error(
                        "packetfilter only supports destination interface XOR source interface per term"
                    )

                term_direction = direction
                if term.destination_interface:
                    term_direction = 'out'
                    term.interface = term.destination_interface
                elif term.source_interface:
                    term_direction = 'in'
                    term.interface = term.source_interface
                else:
                    term.interface = None

                new_terms.append(
                    self._TERM(
                        term, filter_name, all_protocols_stateful, filter_type, term_direction
                    )
                )

            self.pf_policies.append((header, filter_name, filter_type, new_terms))

    def __str__(self) -> str:
        """Render the output of the PF policy into config."""
        target = []
        pretty_platform = f'{self._PLATFORM[0].upper()}{self._PLATFORM[1:]}'
        # Create address table.
        for name in sorted(self.address_book):
            v4 = sorted([x for x in self.address_book[name] if x.version == 4])
            v6 = sorted([x for x in self.address_book[name] if x.version == 6])
            entries = ',\\\n'.join(str(x) for x in v4 + v6)
            target.append(f'table <{name}> {{{entries}}}')
        # pylint: disable=unused-variable
        for header, filter_name, filter_type, terms in self.pf_policies:
            # Add comments for this filter
            target.append(f'# {pretty_platform} {header.FilterName(self._PLATFORM)} Policy')

            # reformat long text comments, if needed
            comments = aclgenerator.WrapWords(header.comment, 70)
            if comments and comments[0]:
                for line in comments:
                    target.append(f'# {line}')
                target.append('#')
            # add the p4 tags
            target.extend(aclgenerator.AddRepositoryTags('# '))
            target.append(f"# {filter_type}")

            # add the terms
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.append(term_str)
            target.append('')

        return '\n'.join(target)



import collections
import heapq
import ipaddress
import itertools
from typing import Union

from aerleon.lib.fqdn import FQDN
from aerleon.lib.nacaddr import IPv4, IPv6


class Addressbook:
    def __init__(self) -> None:
        self._fqdn_addressbook = collections.OrderedDict()
        self._ip_addressbook = collections.OrderedDict()

    def GetZoneNames(self) -> list:
        """Get each zone name contained within the address book.

        Returns:
          A list containing the names of each zone within the addressbook.
        """
        return list(self._ip_addressbook.keys())

    def Walk(self, zone=None):
        """Walks the addressbook by each zone and address/fqdns within.

        Args:
          zone_filter: A specific zone to walk rather than the entire addressbook.

        Yields:
          izone: The zone of the addressbook.
          group: The token of a set of address objects.
          ips: A list of nacaddr objects.
          fqdns: A list of FQDN objects.
        """
        for izone in self._ip_addressbook:
            if zone and izone != zone:
                continue
            groups = set(self._ip_addressbook[izone].keys())
            groups.update(list(self._fqdn_addressbook[izone].keys()))
            for group in sorted(groups):
                ips = []
                fqdns = []
                if group in self._ip_addressbook[izone]:
                    ips = self._ip_addressbook[izone][group]
                if group in self._fqdn_addressbook[izone]:
                    fqdns = self._fqdn_addressbook[izone][group]
                yield izone, group, ips, fqdns

    def GetFQDN(self, zone: str, name: str):
        """Gets an FQDN entry from the addressbook.
        Args:
          zone: The zone to retrieve from.
          name: The name of the address object to retrieve.
        Returns:
          A list of FQDN objects.
        """
        return self._fqdn_addressbook[zone][name]

    def GetAddressTokensInZone(self, zone: str):
        """Gets all address tokens in a zone.
        Args:
          zone: The zone to retrieve from.
        Returns:
          A list of address tokens."""
        return [i for i in self._ip_addressbook[zone]]

    def GetFQDNTokensInZone(self, zone: str):
        """Gets all FQDN tokens in a zone.
        Args:
          zone: The zone to retrieve from.
        Returns:
          A list of FQDN tokens
        """
        return [i for i in self._fqdn_addressbook[zone]]

    def AddFQDNs(self, zone: str, fqdn_list: list[FQDN], suffix=''):
        """Adds a list of FQDNs to the addressbook.
        Args:
          zone: The zone in which to add the FQDNs.
          fqdn_list: A list of FQDN objects.
          suffix: An optional suffix to add to the token.
        """
        if zone not in self._fqdn_addressbook:
            self._ip_addressbook[zone] = collections.defaultdict(list)
            self._fqdn_addressbook[zone] = collections.defaultdict(list)
        for fqdn in fqdn_list:
            self._fqdn_addressbook[zone][fqdn.parent_token + suffix].append(fqdn)

    def GetAddress(self, zone: str, name: str):
        """Gets an Nacaddr entry from the addressbook.
        Args:
          zone: The zone to retrieve from.
          name: The name of the address object to retrieve.
        Returns:
          A list of Nacaddr objects.
        """
        return self._ip_addressbook[zone][name]

    def AddAddresses(self, zone: str, address_list: list[Union[IPv4, IPv6]]):
        """Create the address book configuration entries.

        Args:
          zone: the zone these objects will reside in
          address_list: a list of naming library address objects
            that will reside in the zone

        _BuildAddressBook will add the given addresses to the address book
        for 'zone', grouped by address.parent_name. It will ignore any redundant networks.
        A redundant network is fully contained by an equal or larger network in the list.
        A smaller network already present in the address book will be removed if made redundant
        when new networks are added.
        """

        def _drop_subnets(address_list: list[Union[IPv4, IPv6]]):
            """Remove any network contained by another network in this list.

            Args:
                address_list: a list of IP objects sorted ascending by version and address.

            Yields:
                yields the items of the address list in order with redundant subnets removed.

            _drop_subnets uses the following strategy to identify and discard redundant subnets:
            (1) Assume address_list is a list of IPv{4,6} networks.
                (1A) The address_list must be sorted by (
                    address.version,
                    address.network_address,
                    address.netmask).
            (2) Partition the list by IP version.
            (3) Each partition is sorted by network_address (ascending) and netmask (ascending).
            (4) Network_address gives the low end of the network IP range and broadcast_address gives the
                high end. Broadcast address is always calculated as network_address | netmask.
            (5) A network is a subnet of another network if and only if it has broacast_address less than or equal
                to the highest broadcast address seen thus far. All networks seen thus far have network_address
                lower than or equal to network_address (the low end), so any time broadcast_address (the high end) is
                lower than or equal to the previous max we have definitely found a redundant subnet. And because all
                networks seen thus far have network_address lower than or equal to network_address AND a lower netmask,
                no previous network can be a subnet of the current network (unless they are exactly equal).
            """
            # Partition by IP version: address_list may contain mixed IPv4 / IPv6 addresses
            for _, address_list in itertools.groupby(
                address_list, key=lambda address: address.version
            ):
                last_broadcast_addr = None
                for address in address_list:
                    if (
                        last_broadcast_addr is not None
                        and address.broadcast_address <= last_broadcast_addr
                    ):
                        continue
                    last_broadcast_addr = address.broadcast_address
                    yield address

        if zone not in self._ip_addressbook:
            self._ip_addressbook[zone] = collections.defaultdict(list)
            self._fqdn_addressbook[zone] = collections.defaultdict(list)

        # sort by (parent_token, version, address),
        # then partition by parent_token
        # the implicit sort key for address (always a network) is (version, network_address, netmask)
        for parent_token, address_list in itertools.groupby(
            sorted(
                address_list,
                key=lambda address: (
                    address.parent_token,
                    ipaddress.get_mixed_type_key(address),
                ),
            ),
            key=lambda address: address.parent_token,
        ):
            # merge sorted lists of IP objects
            self._ip_addressbook[zone][parent_token] = list(
                heapq.merge(
                    self._ip_addressbook[zone][parent_token],
                    address_list,
                    key=ipaddress.get_mixed_type_key,
                )
            )

            # drop redundant addresses and networks
            self._ip_addressbook[zone][parent_token] = list(
                _drop_subnets(self._ip_addressbook[zone][parent_token])
            )



# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Google Compute Engine firewall generator.

More information about GCE networking and firewalls:
https://cloud.google.com/compute/docs/networking
https://cloud.google.com/compute/docs/reference/latest/firewalls
"""

import copy
import ipaddress
import json
import re
from typing import TypedDict, Union

from absl import logging

from aerleon.lib import gcp, nacaddr, policy


class Error(gcp.Error):
    """Generic error class."""


class GceFirewallError(Error):
    """Raised with problems in formatting for GCE firewall."""


class ExceededAttributeCountError(Error):
    """Raised when the total attribute count of a policy is above the maximum."""


class LogConfig(TypedDict):
    enable: bool


class L4Matcher(TypedDict):
    IPProtocol: str
    ports: 'list[str]'


class FirewallRule(TypedDict):
    name: str
    description: str
    network: str
    priority: int
    sourceRanges: 'list[str]'
    destinationRanges: 'list[str]'
    sourceTags: 'list[str]'
    targetTags: 'list[str]'
    allowed: 'list[L4Matcher]'
    denied: 'list[L4Matcher]'
    direction: str  # Actually Enum
    logConfig: LogConfig


def IsDefaultDeny(term: policy.Term) -> bool:
    """Returns true if a term is a default deny without IPs, ports, etc."""
    skip_attrs = [
        'flattened',
        'flattened_addr',
        'flattened_saddr',
        'flattened_daddr',
        'action',
        'comment',
        'name',
        'logging',
    ]
    if 'deny' not in term.action:
        return False
    # This lc will look through all methods and attributes of the object.
    # It returns only the attributes that need to be looked at to determine if
    # this is a default deny.
    for i in [
        a
        for a in dir(term)
        if not a.startswith('__') and a.islower() and not callable(getattr(term, a))
    ]:
        if i in skip_attrs:
            continue
        v = getattr(term, i)
        if isinstance(v, str) and v:
            return False
        if isinstance(v, list) and v:
            return False

    return True


def GetNextPriority(priority: int) -> int:
    """Get the priority for the next rule."""
    return priority


class Term(gcp.Term):
    """Creates the term for the GCE firewall."""

    ACTION_MAP = {'accept': 'allowed', 'deny': 'denied'}
    # Restrict the number of addresses per term to 256.
    # Similar restrictions apply to source and target tags, and ports.
    # Details: https://cloud.google.com/vpc/docs/quota#per_network_2
    _TERM_ADDRESS_LIMIT = 256
    _TERM_SOURCE_TAGS_LIMIT = 30
    _TERM_TARGET_TAGS_LIMIT = 70
    _TERM_PORTS_LIMIT = 256

    # Firewall rule name has to match specific RE:
    # The first character must be a lowercase letter, and all following characters
    # must be a dash, lowercase letter, or digit, except the last character, which
    # cannot be a dash.
    # Details: https://cloud.google.com/compute/docs/reference/latest/firewalls
    _TERM_NAME_RE = re.compile(r'^[a-z]([-a-z0-9]*[a-z0-9])?$')

    # Protocols allowed by name from:
    # https://cloud.google.com/vpc/docs/firewalls#protocols_and_ports
    _ALLOW_PROTO_NAME = frozenset(
        [
            'tcp',
            'udp',
            'icmp',
            'esp',
            'ah',
            'ipip',
            'sctp',
            'all',  # Needed for default deny, do not use in policy file.
        ]
    )

    # Any protocol not in _ALLOW_PROTO_NAME must be passed by number.
    ALWAYS_PROTO_NUM = set(gcp.Term.PROTO_MAP.keys()) - _ALLOW_PROTO_NAME

    def __init__(
        self, term: policy.Term, inet_version: str = 'inet', policy_inet_version: str = 'inet'
    ) -> None:
        super().__init__(term)
        self.term = term
        self.inet_version = inet_version
        # This is to handle mixed, where the policy_inet_version is mixed,
        # but the term inet version is either inet/inet6.
        # This is only useful for term name and priority.
        self.policy_inet_version = policy_inet_version

        self._validateDirection()
        if self.term.source_address_exclude and not self.term.source_address:
            raise GceFirewallError(
                'GCE firewall does not support address exclusions without a source '
                'address list.'
            )
        # The reason for the error below isn't because of a GCE restriction, but
        # because we don't want to use a bad default of GCE that allows talking
        # to anything when there's no source address, source tag, or source service
        # account.
        if (
            not self.term.source_address and not self.term.source_tag
        ) and self.term.direction == 'INGRESS':
            raise GceFirewallError(
                'GCE firewall needs either to specify source address or source tags.'
            )
        if self.term.source_port:
            raise GceFirewallError('GCE firewall does not support source port restrictions.')
        if (
            self.term.source_address_exclude
            and self.term.source_address
            or self.term.destination_address_exclude
            and self.term.destination_address
        ):
            self.term.FlattenAll()
            if not self.term.source_address and self.term.direction == 'INGRESS':
                raise GceFirewallError(
                    'GCE firewall rule no longer contains any source addresses after '
                    'the prefixes in source_address_exclude were removed.'
                )
            # Similarly to the comment above, the reason for this error is also
            # because we do not want to use the bad default of GCE that allows for
            # talking to anything when there is no IP address provided for this field.
            if not self.term.destination_address and self.term.direction == 'EGRESS':
                raise GceFirewallError(
                    'GCE firewall rule no longer contains any destination addresses '
                    'after the prefixes in destination_address_exclude were removed.'
                )

    def __str__(self):
        """Convert term to a string."""
        json.dumps(self.ConvertToDict(), indent=2, separators=(',', ': '))

    def _validateDirection(self) -> None:
        if self.term.direction == 'INGRESS':
            if not self.term.source_address and not self.term.source_tag:
                raise GceFirewallError(
                    'Ingress rule missing required field oneof "sourceRanges" or ' '"sourceTags".'
                )

            if self.term.destination_address:
                raise GceFirewallError('Ingress rules cannot include ' '"destinationRanges.')

        elif self.term.direction == 'EGRESS':
            if self.term.source_address:
                raise GceFirewallError('Egress rules cannot include "sourceRanges".')

            if not self.term.destination_address:
                raise GceFirewallError('Egress rule missing required field "destinationRanges".')

            if self.term.destination_tag:
                raise GceFirewallError('GCE Egress rule cannot have destination tag.')

    def ConvertToDict(self) -> list[FirewallRule]:
        """Convert term to a dictionary.

        This is used to get a dictionary describing this term which can be
        output easily as a JSON blob.

        Returns:
          A dictionary that contains all fields necessary to create or update a GCE
          firewall.

        Raises:
          GceFirewallError: The term name is too long.
        """
        if self.term.owner:
            self.term.comment.append(f'Owner: {self.term.owner}')
        term_dict = {
            'description': ' '.join(self.term.comment),
            'name': self.term.name,
            'direction': self.term.direction,
        }
        if self.term.network:
            term_dict['network'] = self.term.network
            term_dict['name'] = f"{self.term.network.split('/')[-1]}-{term_dict['name']}"
        # Identify if this is inet6 processing for a term under a mixed policy.
        mixed_policy_inet6_term = False
        if self.policy_inet_version == 'mixed' and self.inet_version == 'inet6':
            mixed_policy_inet6_term = True
        # Update term name to have the IPv6 suffix for the inet6 rule.
        if mixed_policy_inet6_term:
            term_dict['name'] = gcp.GetIpv6TermName(term_dict['name'])

        # Checking counts of tags, and ports to see if they exceeded limits.
        if len(self.term.source_tag) > self._TERM_SOURCE_TAGS_LIMIT:
            raise GceFirewallError(
                f'GCE firewall rule exceeded number of source tags per rule: {self.term.name}'
            )
        if len(self.term.destination_tag) > self._TERM_TARGET_TAGS_LIMIT:
            raise GceFirewallError(
                f'GCE firewall rule exceeded number of target tags per rule: {self.term.name}'
            )

        if self.term.source_tag:
            if self.term.direction == 'INGRESS':
                term_dict['sourceTags'] = self.term.source_tag
            elif self.term.direction == 'EGRESS':
                term_dict['targetTags'] = self.term.source_tag
        if self.term.destination_tag and self.term.direction == 'INGRESS':
            term_dict['targetTags'] = self.term.destination_tag
        if self.term.priority:
            term_dict['priority'] = self.term.priority
            # Update term priority for the inet6 rule.
            if mixed_policy_inet6_term:
                term_dict['priority'] = GetNextPriority(term_dict['priority'])

        rules = []
        # If 'mixed' ends up in indvidual term inet_version, something has gone
        # horribly wrong. The only valid values are inet/inet6.
        term_af = self.AF_MAP.get(self.inet_version)
        if self.inet_version == 'mixed':
            raise GceFirewallError(
                f'GCE firewall rule has incorrect inet_version for rule: {self.term.name}'
            )

        # Exit early for inet6 processing of mixed rules that have only tags,
        # and no IP addresses, since this is handled in the inet processing.
        if mixed_policy_inet6_term:
            if not self.term.source_address and not self.term.destination_address:
                if 'targetTags' in term_dict or 'sourceTags' in term_dict:
                    return []

        saddrs = sorted(
            self.term.GetAddressOfVersion('source_address', term_af),
            key=ipaddress.get_mixed_type_key,
        )
        daddrs = sorted(
            self.term.GetAddressOfVersion('destination_address', term_af),
            key=ipaddress.get_mixed_type_key,
        )

        # If the address got filtered out and is empty due to address family, we
        # don't render the term. At this point of term processing, the direction
        # has already been validated, so we can just log and return empty rule.
        if self.term.source_address and not saddrs:
            logging.warning(
                'WARNING: Term %s is not being rendered for %s, '
                'because there are no addresses of that family.',
                self.term.name,
                self.inet_version,
            )
            return []
        if self.term.destination_address and not daddrs:
            logging.warning(
                'WARNING: Term %s is not being rendered for %s, '
                'because there are no addresses of that family.',
                self.term.name,
                self.inet_version,
            )
            return []

        filtered_protocols = []
        if not self.term.protocol:
            # Any protocol is represented as "all"
            filtered_protocols = ['all']
            logging.info(
                'INFO: Term %s has no protocol specified,'
                'which is interpreted as "all" protocols.',
                self.term.name,
            )

        proto_dict = copy.deepcopy(term_dict)

        if self.term.logging:
            proto_dict['logConfig'] = {'enable': True}

        for proto in self.term.protocol:
            # ICMP filtering by inet_version
            # Since each term has inet_version, 'mixed' is correctly processed here.
            # Convert protocol to number for uniformity of comparison.
            # PROTO_MAP always returns protocol number.
            if proto in self._ALLOW_PROTO_NAME:
                proto_num = self.PROTO_MAP[proto]
            else:
                proto_num = proto
            if proto_num == self.PROTO_MAP['icmp'] and self.inet_version == 'inet6':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet6, ICMP '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            if proto_num == self.PROTO_MAP['icmpv6'] and self.inet_version == 'inet':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet, ICMPv6 '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            if proto_num == self.PROTO_MAP['igmp'] and self.inet_version == 'inet6':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet6, IGMP '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            filtered_protocols.append(proto)
        # If there is no protocol left after ICMP/IGMP filtering, drop this term.
        if not filtered_protocols:
            return []
        for proto in filtered_protocols:
            # If the protocol name is not supported, protocol number is used.
            # This is done by default in policy.py.
            if proto not in self._ALLOW_PROTO_NAME:
                logging.info(
                    'INFO: Term %s is being rendered using protocol number', self.term.name
                )
            dest = {'IPProtocol': proto}

            if self.term.destination_port:
                ports = []
                for start, end in self.term.destination_port:
                    if start == end:
                        ports.append(str(start))
                    else:
                        ports.append('%d-%d' % (start, end))
                if len(ports) > self._TERM_PORTS_LIMIT:
                    raise GceFirewallError(
                        f'GCE firewall rule exceeded number of ports per rule: {self.term.name}'
                    )
                dest['ports'] = ports

            action = self.ACTION_MAP[self.term.action[0]]
            dict_val = []
            if action in proto_dict:
                dict_val = proto_dict[action]
                if not isinstance(dict_val, list):
                    dict_val = [dict_val]
            dict_val.append(dest)
            proto_dict[action] = dict_val

        # There's a limit of 256 addresses each term can contain.
        # If we're above that limit, we're breaking it down in more terms.
        if saddrs:
            source_addr_chunks = [
                saddrs[x : x + self._TERM_ADDRESS_LIMIT]
                for x in range(0, len(saddrs), self._TERM_ADDRESS_LIMIT)
            ]
            for i, chunk in enumerate(source_addr_chunks):
                rule = copy.deepcopy(proto_dict)
                if len(source_addr_chunks) > 1:
                    rule['name'] = '%s-%d' % (rule['name'], i + 1)
                rule['sourceRanges'] = [str(saddr) for saddr in chunk]
                rules.append(rule)
        elif daddrs:
            dest_addr_chunks = [
                daddrs[x : x + self._TERM_ADDRESS_LIMIT]
                for x in range(0, len(daddrs), self._TERM_ADDRESS_LIMIT)
            ]
            for i, chunk in enumerate(dest_addr_chunks):
                rule = copy.deepcopy(proto_dict)
                if len(dest_addr_chunks) > 1:
                    rule['name'] = '%s-%d' % (rule['name'], i + 1)
                rule['destinationRanges'] = [str(daddr) for daddr in chunk]
                rules.append(rule)
        else:
            rules.append(proto_dict)

        # Sanity checking term name lengths.
        long_rules = [rule['name'] for rule in rules if len(rule['name']) > 63]
        if long_rules:
            raise GceFirewallError(f'GCE firewall name ended up being too long: {long_rules}')
        return rules


class GCE(gcp.GCP):
    """A GCE firewall policy object."""

    _PLATFORM = 'gce'
    SUFFIX = '.gce'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'mixed'))
    _ANY_IP = {
        'inet': nacaddr.IP('0.0.0.0/0'),
        'inet6': nacaddr.IP('::/0'),
    }
    # Supported is 63 but we need to account for dynamic updates when the term
    # is rendered (which can add proto and a counter).
    _TERM_MAX_LENGTH = 53
    _GOOD_DIRECTION = ['INGRESS', 'EGRESS']
    _OPTIONAL_SUPPORTED_KEYWORDS = {'expiration', 'destination_tag', 'source_tag'}

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, _ = super()._BuildTokens()

        # add extra things
        supported_tokens |= {'destination_tag', 'expiration', 'owner', 'priority', 'source_tag'}

        # remove unsupported things
        supported_tokens -= {'icmp_type', 'verbatim'}
        # easier to make a new structure
        supported_sub_tokens = {'action': {'accept', 'deny'}}

        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        self.gce_policies = []
        max_attribute_count = 0
        total_attribute_count = 0
        total_rule_count = 0

        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            network = ''
            direction = 'INGRESS'
            if filter_options:
                for i in self._GOOD_DIRECTION:
                    if i in filter_options:
                        direction = i
                        filter_options.remove(i)
            # Get the address family if set.
            address_family = 'inet'
            for i in self._SUPPORTED_AF:
                if i in filter_options:
                    address_family = i
                    filter_options.remove(i)

            for opt in filter_options:
                try:
                    max_attribute_count = int(opt)
                    logging.info('Checking policy for max attribute count %d', max_attribute_count)
                    filter_options.remove(opt)
                    break
                except ValueError:
                    continue

            if filter_options:
                network = filter_options[0]
            else:
                logging.warning('GCE filter does not specify a network.')

            if not terms:
                continue

            term_names = set()
            if IsDefaultDeny(terms[-1]):
                terms[-1].protocol = ['all']
                terms[-1].priority = 65534
                if direction == 'EGRESS':
                    if address_family != 'mixed':
                        # Default deny also gets processed as part of terms processing.
                        # The name and priority get updated there.
                        terms[-1].destination_address = [self._ANY_IP[address_family]]
                    else:
                        terms[-1].destination_address = [
                            self._ANY_IP['inet'],
                            self._ANY_IP['inet6'],
                        ]
                else:
                    if address_family != 'mixed':
                        terms[-1].source_address = [self._ANY_IP[address_family]]
                    else:
                        terms[-1].source_address = [self._ANY_IP['inet'], self._ANY_IP['inet6']]

            for term in terms:
                if term.stateless_reply:
                    logging.warning(
                        'WARNING: Term %s in policy %s is a stateless reply '
                        'term and will not be rendered.',
                        term.name,
                        filter_name,
                    )
                    continue
                term.network = network
                if not term.comment:
                    term.comment = header.comment
                if direction == 'EGRESS':
                    term.name += '-e'
                term.name = self.FixTermLength(term.name)
                if term.name in term_names:
                    raise GceFirewallError(f'Duplicate term name {term.name}')
                term_names.add(term.name)

                term.direction = direction
                if term.option:
                    raise GceFirewallError('GCE firewall does not support term options.')

                # Handle mixed for each indvidual term as inet and inet6.
                # inet/inet6 are treated the same.
                term_address_families = []
                if address_family == 'mixed':
                    term_address_families = ['inet', 'inet6']
                else:
                    term_address_families = [address_family]
                for term_af in term_address_families:
                    for rules in Term(term, term_af, address_family).ConvertToDict():
                        logging.debug(
                            'Attribute count of rule %s is: %d',
                            term.name,
                            GetAttributeCount(rules),
                        )
                        total_attribute_count += GetAttributeCount(rules)
                        total_rule_count += 1
                        if max_attribute_count and total_attribute_count > max_attribute_count:
                            # Stop processing rules as soon as the attribute count is over the
                            # limit.
                            raise ExceededAttributeCountError(
                                'Attribute count (%d) for %s exceeded the maximum (%d)'
                                % (total_attribute_count, filter_name, max_attribute_count)
                            )
                        self.gce_policies.append(rules)
        logging.info('Total rule count of policy %s is: %d', filter_name, total_rule_count)
        logging.info(
            'Total attribute count of policy %s is: %d', filter_name, total_attribute_count
        )

    def __str__(self) -> str:
        out = '%s\n\n' % (
            json.dumps(self.gce_policies, indent=2, separators=(',', ': '), sort_keys=True)
        )

        return out


def GetAttributeCount(dict_term: dict[str, Union[list, str]]) -> int:
    """Calculate the attribute count of a term in its dictionary form.

    The attribute count of a rule is the sum of the number of ports, protocols, IP
    ranges, tags and target service account.

    Note: The goal of this function is not to determine if a term is valid, but
        to calculate its attribute count regardless of correctness.

    Args:
      dict_term: A dict object.

    Returns:
      int: The attribute count of the term.
    """
    addresses = len(dict_term.get('destinationRanges', [])) or len(
        dict_term.get('sourceRanges', [])
    )

    proto_ports = 0
    for allowed in dict_term.get('allowed', []):
        proto_ports += len(allowed.get('ports', [])) + 1  # 1 for ipProtocol
    for denied in dict_term.get('denied', []):
        proto_ports += len(denied.get('ports', [])) + 1  # 1 for ipProtocol

    tags = 0
    for _ in dict_term.get('sourceTags', []):
        tags += 1
    for _ in dict_term.get('targetTags', []):
        tags += 1

    service_accounts = 0
    for _ in dict_term.get('targetServiceAccount', []):
        service_accounts += 1

    return addresses + proto_ports + tags + service_accounts
